# -*- coding: utf-8 -*-
"""
This module needs serious refactoring and testing
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import shelve
import six
import uuid
import json
import codecs
import os
#import lru
#git+https://github.com/amitdev/lru-dict
#import atexit
#import inspect
import contextlib
import collections
from six.moves import cPickle as pickle
from six.moves import range, zip
from os.path import join, normpath, basename, exists
from functools import partial
from itertools import chain
import zipfile
from utool import util_arg
from utool import util_hash
from utool import util_inject
from utool import util_path
from utool import util_io
from utool import util_str
from utool import util_cplat
from utool import util_inspect
from utool import util_list
from utool import util_class
from utool import util_type
from utool import util_decor  # NOQA
from utool import util_dict
from utool._internal import meta_util_constants
print, rrr, profile = util_inject.inject2(__name__)


# TODO: Remove globalness

VERBOSE = util_arg.VERBOSE
QUIET = util_arg.QUIET
VERBOSE_CACHE = util_arg.NOT_QUIET
USE_CACHE = not util_arg.get_argflag('--nocache')
__APPNAME__ = meta_util_constants.default_appname  # the global application name


class CacheMissException(Exception):
    pass


#class YACacher(object):
# @six.add_metaclass(util_class.ReloadingMetaclass)
@util_class.reloadable_class
class ShelfCacher(object):
    """ yet another cacher """
    def __init__(self, fpath, enabled=True):
        self.verbose = True
        if self.verbose:
            print('[shelfcache] initializing()')
        self.fpath = fpath
        self.shelf = None if not enabled else shelve.open(fpath)

    def __del__(self):
        self.close()

    def __getitem__(self, cachekey):
        return self.load(cachekey)

    def __setitem__(self, cachekey, data):
        return self.save(cachekey, data)

    def keys(self):
        return self.shelf.keys()

    def load(self, cachekey):
        if self.verbose:
            print('[shelfcache] loading %s' % (cachekey,))

        cachekey = cachekey.encode('ascii')
        if self.shelf is None or cachekey not in self.shelf:
            raise CacheMissException(
                'Cache miss cachekey=%r self.fpath=%r' % (cachekey, self.fpath))
        else:
            return self.shelf[cachekey]

    def save(self, cachekey, data):
        if self.verbose:
            print('[shelfcache] saving %s' % (cachekey,))

        cachekey = cachekey.encode('ascii')
        if self.shelf is not None:
            self.shelf[cachekey] = data
        self.shelf.sync()

    def clear(self):
        if self.verbose:
            print('[shelfcache] clearing cache')
        self.shelf.clear()
        self.shelf.sync()

    def close(self):
        if self.verbose:
            print('[shelfcache] closing()')
        if self.shelf is not None:
            self.shelf.close()


def get_default_appname():
    global __APPNAME__
    return __APPNAME__


def text_dict_read(fpath):
    try:
        with open(fpath, 'r') as file_:
            dict_text = file_.read()
    except IOError:
        dict_text = '{}'
    try:
        dict_ = eval(dict_text, {}, {})
    except SyntaxError as ex:
        import utool as ut
        print(dict_text)
        ut.printex(ex, 'Bad Syntax', keys=['dict_text'])
        dict_ = {}
        if util_arg.SUPER_STRICT:
            raise
    return dict_


def text_dict_write(fpath, dict_):
    """
    Very naive, but readable way of storing a dictionary on disk
    FIXME: This broke on RoseMary's big dataset. Not sure why. It gave bad
    syntax. And the SyntaxError did not seem to be excepted.
    """
    #dict_ = text_dict_read(fpath)
    #dict_[key] = val
    dict_text2 = util_str.repr4(dict_, strvals=False)
    if VERBOSE:
        print('[cache] ' + str(dict_text2))
    util_io.write_to(fpath, dict_text2)


def consensed_cfgstr(prefix, cfgstr, max_len=128, cfgstr_hashlen=16):
    if len(prefix) + len(cfgstr) > max_len:
        hashed_cfgstr = util_hash.hashstr27(cfgstr, hashlen=cfgstr_hashlen)
        # Hack for prettier names
        if not prefix.endswith('_'):
            fname_cfgstr = prefix + '_' + hashed_cfgstr
        else:
            fname_cfgstr = prefix + hashed_cfgstr
    else:
        fname_cfgstr = prefix + cfgstr
    return fname_cfgstr


def _args2_fpath(dpath, fname, cfgstr, ext):
    r"""
    Ensures that the filename is not too long

    Internal util_cache helper function
    Windows MAX_PATH=260 characters
    Absolute length is limited to 32,000 characters
    Each filename component is limited to 255 characters

    Args:
        dpath (str):
        fname (str):
        cfgstr (str):
        ext (str):

    Returns:
        str: fpath

    CommandLine:
        python -m utool.util_cache --test-_args2_fpath

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_cache import *  # NOQA
        >>> from utool.util_cache import _args2_fpath
        >>> import utool as ut
        >>> dpath = 'F:\\data\\work\\PZ_MTEST\\_ibsdb\\_ibeis_cache'
        >>> fname = 'normalizer_'
        >>> cfgstr = u'PZ_MTEST_DSUUIDS((9)67j%dr%&bl%4oh4+)_QSUUIDS((9)67j%dr%&bl%4oh4+)zebra_plains_vsone_NN(single,K1+1,last,cks1024)_FILT(ratio<0.625;1.0,fg;1.0)_SV(0.01;2;1.57minIn=4,nRR=50,nsum,)_AGG(nsum)_FLANN(4_kdtrees)_FEATWEIGHT(ON,uselabel,rf)_FEAT(hesaff+sift_)_CHIP(sz450)'
        >>> ext = '.cPkl'
        >>> fpath = _args2_fpath(dpath, fname, cfgstr, ext)
        >>> result = str(ut.ensure_unixslash(fpath))
        >>> target = 'F:/data/work/PZ_MTEST/_ibsdb/_ibeis_cache/normalizer_xfylfboirymmcpfg.cPkl'
        >>> ut.assert_eq(result, target)

    """
    if len(ext) > 0 and ext[0] != '.':
        raise ValueError('Please be explicit and use a dot in ext')
    max_len = 128
    # should hashlen be larger?
    cfgstr_hashlen = 16
    prefix = fname
    fname_cfgstr = consensed_cfgstr(prefix, cfgstr, max_len=max_len,
                                    cfgstr_hashlen=cfgstr_hashlen)
    fpath = join(dpath, fname_cfgstr + ext)
    fpath = normpath(fpath)
    return fpath


def save_cache(dpath, fname, cfgstr, data, ext='.cPkl', verbose=None):
    """
    Saves data using util_io, but smartly constructs a filename
    """
    fpath = _args2_fpath(dpath, fname, cfgstr, ext)
    util_io.save_data(fpath, data, verbose=verbose)
    return fpath


def load_cache(dpath, fname, cfgstr, ext='.cPkl', verbose=None, enabled=True):
    """
    Loads data using util_io, but smartly constructs a filename
    """
    if verbose is None:
        verbose = VERBOSE_CACHE
    if not USE_CACHE or not enabled:
        if verbose > 1:
            print('[util_cache] ... cache disabled: dpath=%s cfgstr=%r' %
                    (basename(dpath), cfgstr,))
        raise IOError(3, 'Cache Loading Is Disabled')
    fpath = _args2_fpath(dpath, fname, cfgstr, ext)
    if not exists(fpath):
        if verbose > 0:
            print('[util_cache] ... cache does not exist: dpath=%r fname=%r cfgstr=%r' % (
                basename(dpath), fname, cfgstr,))
        raise IOError(2, 'No such file or directory: %r' % (fpath,))
    else:
        if verbose > 2:
            print('[util_cache] ... cache exists: dpath=%r fname=%r cfgstr=%r' % (
                basename(dpath), fname, cfgstr,))
        import utool as ut
        nbytes = ut.get_file_nBytes(fpath)
        big_verbose = (nbytes > 1E6 and verbose > 2) or verbose > 2
        if big_verbose:
            print('[util_cache] About to read file of size %s' % (ut.byte_str2(nbytes),))
    try:
        with ut.Timer(fpath, verbose=big_verbose and verbose > 3):
            data = util_io.load_data(fpath, verbose=verbose > 2)
    except (EOFError, IOError, ImportError) as ex:
        print('CORRUPTED? fpath = %s' % (fpath,))
        if verbose > 1:
            print('[util_cache] ... cache miss dpath=%s cfgstr=%r' % (
                basename(dpath), cfgstr,))
        raise IOError(str(ex))
    except Exception:
        print('CORRUPTED? fpath = %s' % (fpath,))
        raise
    else:
        if verbose > 2:
            print('[util_cache] ... cache hit')
    return data


def tryload_cache(dpath, fname, cfgstr, verbose=None):
    """
    returns None if cache cannot be loaded
    """
    try:
        return load_cache(dpath, fname, cfgstr, verbose=verbose)
    except IOError:
        return None


@profile
def tryload_cache_list(dpath, fname, cfgstr_list, verbose=False):
    """
    loads a list of similar cached datas. Returns flags that needs to be computed
    """
    data_list = [tryload_cache(dpath, fname, cfgstr, verbose) for cfgstr in cfgstr_list]
    ismiss_list = [data is None for data in data_list]
    return data_list, ismiss_list


@profile
def tryload_cache_list_with_compute(use_cache, dpath, fname, cfgstr_list,
                                    compute_fn, *args):
    """
    tries to load data, but computes it if it can't give a compute function
    """
    # Load precomputed values
    if use_cache is False:
        data_list = [None] * len(cfgstr_list)
        ismiss_list = [True] * len(cfgstr_list)
        # Don't load or save, just compute
        data_list = compute_fn(ismiss_list, *args)
        return data_list
    else:
        data_list, ismiss_list = tryload_cache_list(dpath, fname, cfgstr_list,
                                                    verbose=False)
    num_total = len(cfgstr_list)
    if any(ismiss_list):
        # Compute missing values
        newdata_list = compute_fn(ismiss_list, *args)
        newcfgstr_list = util_list.compress(cfgstr_list, ismiss_list)
        index_list = util_list.list_where(ismiss_list)
        print('[cache] %d/%d cache hits for %s in %s' % (num_total -
                                                         len(index_list),
                                                         num_total, fname,
                                                         util_path.tail(dpath)))
        # Cache write
        for newcfgstr, newdata in zip(newcfgstr_list, newdata_list):
            save_cache(dpath, fname, newcfgstr, newdata, verbose=False)
        # Populate missing result
        for index, newdata in zip(index_list, newdata_list):
            data_list[index] = newdata
    else:
        print('[cache] %d/%d cache hits for %s in %s' % (num_total, num_total,
                                                         fname,
                                                         util_path.tail(dpath)))
    return data_list


class Cacher(object):
    """
    old non inhertable version of cachable
    """
    def __init__(self, fname, cfgstr=None, cache_dir='default',
                 appname='utool', ext='.cPkl', verbose=None,
                 enabled=True):
        if verbose is None:
            verbose = VERBOSE
        if cache_dir == 'default':
            cache_dir = util_cplat.get_app_resource_dir(appname)
        util_path.ensuredir(cache_dir)
        self.dpath = cache_dir
        self.fname = fname
        self.cfgstr = cfgstr
        self.verbose = verbose
        self.ext = ext
        self.enabled = enabled

    def get_fpath(self):
        fpath = _args2_fpath(self.dpath, self.fname, self.cfgstr, self.ext)
        return fpath

    def existing_versions(self):
        """
        Returns data with different cfgstr values that were previously computed
        with this cacher.
        """
        import glob
        pattern = self.fname + '_*' + self.ext
        for fname in glob.glob1(self.dpath, pattern):
            fpath = join(self.dpath, fname)
            yield fpath

    def exists(self, cfgstr=None):
        return exists(self.get_fpath())

    def load(self, cfgstr=None):
        cfgstr = self.cfgstr if cfgstr is None else cfgstr
        # assert cfgstr is not None, 'must specify cfgstr in constructor or call'
        if cfgstr is None:
            import warnings
            warnings.warn('No cfgstr given in Cacher constructor or call')
            cfgstr = ''
        assert self.fname is not None, 'no fname'
        assert self.dpath is not None, 'no dpath'
        # TODO: use the computed fpath from this object instead
        data = load_cache(self.dpath, self.fname, cfgstr, self.ext,
                          verbose=self.verbose, enabled=self.enabled)
        if self.verbose > 1:
            print('[cache] ... ' + self.fname + ' Cacher hit')
        return data

    def tryload(self, cfgstr=None):
        """
        Like load, but returns None if the load fails
        """
        if cfgstr is None:
            cfgstr = self.cfgstr
        if cfgstr is None:
            import warnings
            warnings.warn('No cfgstr given in Cacher constructor or call')
            cfgstr = ''
        # assert cfgstr is not None, (
        #     'must specify cfgstr in constructor or call')
        if not self.enabled:
            if self.verbose > 0:
                print('[cache] ... %s Cacher disabled' % (self.fname))
            return None
        try:
            if self.verbose > 1:
                print('[cache] tryload fname=%s' % (self.fname,))
                # if self.verbose > 2:
                #     print('[cache] cfgstr=%r' % (cfgstr,))
            return self.load(cfgstr)
        except IOError:
            if self.verbose > 0:
                print('[cache] ... %s Cacher miss' % (self.fname))

    def ensure(self, func, *args, **kwargs):
        data = self.tryload()
        if data is None:
            data = func(*args, **kwargs)
            self.save(data)
        return data

    def save(self, data, cfgstr=None):
        if not self.enabled:
            return
        cfgstr = self.cfgstr if cfgstr is None else cfgstr
        # assert cfgstr is not None, 'must specify cfgstr in constructor or call'
        if cfgstr is None:
            import warnings
            warnings.warn('No cfgstr given in Cacher constructor or call')
            cfgstr = ''
        assert self.fname is not None, 'no fname'
        assert self.dpath is not None, 'no dpath'
        if self.verbose > 0:
            print('[cache] ... ' + self.fname + ' Cacher save')
        save_cache(self.dpath, self.fname, cfgstr, data, self.ext)


#@util_decor.memoize
def make_utool_json_encoder(allow_pickle=False):
    """
    References:
        http://stackoverflow.com/questions/8230315/python-sets-are
        http://stackoverflow.com/questions/11561932/why-does-json
        https://github.com/jsonpickle/jsonpickle
        http://stackoverflow.com/questions/24369666/typeerror-b1
        http://stackoverflow.com/questions/30469575/how-to-pickle
    """
    import utool as ut
    PYOBJECT_TAG = '__PYTHON_OBJECT__'
    UUID_TAG = '__UUID__'
    SLICE_TAG = '__SLICE__'

    def decode_pickle(text):
        obj = pickle.loads(codecs.decode(text.encode(), 'base64'))
        return obj

    def encode_pickle(obj):
        try:
            # Use protocol 2 to support both python2.7 and python3
            COMPATIBLE_PROTOCOL = 2
            pickle_bytes = pickle.dumps(obj, protocol=COMPATIBLE_PROTOCOL)
        except Exception:
            raise
        text = codecs.encode(pickle_bytes, 'base64').decode()
        return text

    type_to_tag = collections.OrderedDict([
        (slice, SLICE_TAG),
        (uuid.UUID, UUID_TAG),
        (object, PYOBJECT_TAG),
    ])

    tag_to_type = {tag: type_ for type_, tag in type_to_tag.items()}

    def slice_part(c):
        return '' if c is None else str(c)

    def encode_slice(s):
        parts = [slice_part(s.start), slice_part(s.stop), slice_part(s.step)]
        return ':'.join(parts)

    def decode_slice(x):
        return ut.smart_cast(x, slice)

    encoders = {
        UUID_TAG: str,
        SLICE_TAG: encode_slice,
        PYOBJECT_TAG: encode_pickle,
    }

    decoders = {
        UUID_TAG: uuid.UUID,
        SLICE_TAG: decode_slice,
        PYOBJECT_TAG: decode_pickle,
    }

    if not allow_pickle:
        del encoders[PYOBJECT_TAG]
        del decoders[PYOBJECT_TAG]
        type_ = tag_to_type[PYOBJECT_TAG]
        del tag_to_type[PYOBJECT_TAG]
        del type_to_tag[type_]

    class UtoolJSONEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, util_type.NUMPY_TYPE_TUPLE):
                return obj.tolist()
            elif six.PY3 and isinstance(obj, bytes):
                return obj.decode('utf-8')
            elif isinstance(obj, (set, frozenset)):
                return list(obj)
                # return json.JSONEncoder.default(self, list(obj))
                # return [json.JSONEncoder.default(o) for o in obj]
            elif isinstance(obj, util_type.PRIMATIVE_TYPES):
                return json.JSONEncoder.default(self, obj)
            elif  hasattr(obj, '__getstate__') and not isinstance(obj, uuid.UUID):
                return obj.__getstate__()
            else:
                for type_, tag in type_to_tag.items():
                    if isinstance(obj, type_):
                        #print('----')
                        #print('encoder obj = %r' % (obj,))
                        #print('encoder type_ = %r' % (type_,))
                        func = encoders[tag]
                        text = func(obj)
                        return {tag: text}
                raise TypeError('Invalid serialization type=%r' % (type(obj)))

        @classmethod
        def _json_object_hook(cls, value, verbose=False, **kwargs):
            if len(value) == 1:
                tag, text = list(value.items())[0]
                if tag in decoders:
                    #print('----')
                    #print('decoder tag = %r' % (tag,))
                    func = decoders[tag]
                    obj = func(text)
                    #print('decoder obj = %r' % (obj,))
                    return obj
            else:
                return value
            return value
    return UtoolJSONEncoder


def to_json(val, allow_pickle=False, pretty=False):
    r"""
    Converts a python object to a JSON string using the utool convention

    Args:
        val (object):

    Returns:
        str: json_str

    References:
        http://stackoverflow.com/questions/11561932/why-does-json-dumpslistnp

    CommandLine:
        python -m utool.util_cache --test-to_json
        python3 -m utool.util_cache --test-to_json

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_cache import *  # NOQA
        >>> import utool as ut
        >>> import numpy as np
        >>> import uuid
        >>> val = [
        >>>     '{"foo": "not a dict"}',
        >>>     1.3,
        >>>     [1],
        >>>     # {1: 1, 2: 2, 3: 3}, cant use integer keys
        >>>     {1, 2, 3},
        >>>     slice(1, None, 1),
        >>>     b'an ascii string',
        >>>     np.array([1, 2, 3]),
        >>>     ut.get_zero_uuid(),
        >>>     ut.LazyDict(x='fo'),
        >>>     ut.LazyDict,
        >>>     {'x': {'a', 'b', 'cde'}, 'y': [1]}
        >>> ]
        >>> #val = ut.LazyDict(x='fo')
        >>> allow_pickle = True
        >>> if not allow_pickle:
        >>>     val = val[:-2]
        >>> json_str = ut.to_json(val, allow_pickle=allow_pickle)
        >>> result = ut.repr3(json_str)
        >>> reload_val = ut.from_json(json_str, allow_pickle=allow_pickle)
        >>> # Make sure pickle doesnt happen by default
        >>> try:
        >>>     json_str = ut.to_json(val)
        >>>     assert False or not allow_pickle, 'expected a type error'
        >>> except TypeError:
        >>>     print('Correctly got type error')
        >>> try:
        >>>     json_str = ut.from_json(val)
        >>>     assert False, 'expected a type error'
        >>> except TypeError:
        >>>     print('Correctly got type error')
        >>> print(result)
        >>> print('original = ' + ut.repr3(val, nl=1))
        >>> print('reconstructed = ' + ut.repr3(reload_val, nl=1))
        >>> assert reload_val[6] == val[6].tolist()
        >>> assert reload_val[6] is not val[6]

    Example:
        >>> # test 3.7 safe uuid
        >>> import uuid
        >>> import utool as ut
        >>> ut.to_json([uuid.uuid4()])

    """
    UtoolJSONEncoder = make_utool_json_encoder(allow_pickle)
    json_kw = {}
    json_kw['cls'] = UtoolJSONEncoder
    if pretty:
        json_kw['indent'] = 4
        json_kw['separators'] = (',', ': ')
    json_str = json.dumps(val, **json_kw)
    return json_str


def from_json(json_str, allow_pickle=False):
    """
    Decodes a JSON object specified in the utool convention

    Args:
        json_str (str):
        allow_pickle (bool): (default = False)

    Returns:
        object: val

    CommandLine:
        python -m utool.util_cache from_json --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_cache import *  # NOQA
        >>> import utool as ut
        >>> json_str = 'just a normal string'
        >>> json_str = '["just a normal string"]'
        >>> allow_pickle = False
        >>> val = from_json(json_str, allow_pickle)
        >>> result = ('val = %s' % (ut.repr2(val),))
        >>> print(result)
    """
    if six.PY3:
        if isinstance(json_str, bytes):
            json_str = json_str.decode('utf-8')
    UtoolJSONEncoder = make_utool_json_encoder(allow_pickle)
    object_hook = UtoolJSONEncoder._json_object_hook
    val = json.loads(json_str, object_hook=object_hook)
    return val


def get_func_result_cachekey(func_, args_=tuple(), kwargs_={}):
    """
    TODO: recursive partial definitions
    kwargs = {}
    args = ([],)
    """
    import utool as ut
    # Rectify partials and whatnot
    true_args = args_
    true_kwargs = kwargs_
    true_func = func_
    if isinstance(func_, partial):
        true_func = func_.func
        if func_.args is not None:
            true_args = tuple(list(func_.args) + list(args_))
        if func_.keywords is not None:
            true_kwargs.update(func_.keywords)

    if ut.is_method(true_func):
        method = true_func
        true_func = method.im_func
        self = method.im_self
        true_args = tuple([self] + list(true_args))

    # Build up cachekey
    funcname = ut.get_funcname(true_func)
    kwdefaults = ut.get_kwdefaults(true_func, parse_source=False)
    #kwdefaults = ut.get_kwdefaults(true_func, parse_source=True)
    argnames   = ut.get_argnames(true_func)
    key_argx = None
    key_kwds = None
    func = true_func  # NOQA
    args = true_args  # NOQA
    kwargs = true_kwargs  # NOQA
    args_key = ut.get_cfgstr_from_args(true_func, true_args, true_kwargs,
                                       key_argx, key_kwds, kwdefaults,
                                       argnames)
    cachekey = funcname + '(' + args_key + ')'
    return cachekey


def cachestr_repr(val):
    """
    Representation of an object as a cache string.
    """
    try:
        memview = memoryview(val)
        return memview.tobytes()
    except Exception:
        try:
            return to_json(val)
        except Exception:
            # SUPER HACK
            if repr(val.__class__) == "<class 'ibeis.control.IBEISControl.IBEISController'>":
                return val.get_dbname()


def get_cfgstr_from_args(func, args, kwargs, key_argx, key_kwds, kwdefaults,
                         argnames, use_hash=None):
    """
    Dev:
        argx = ['fdsf', '432443432432', 43423432, 'fdsfsd', 3.2, True]
        memlist = list(map(cachestr_repr, argx))

    Ignore:
        argx = key_argx[0]
        argval = args[argx]
        val = argval
        %timeit repr(argval)
        %timeit to_json(argval)
        %timeit utool.hashstr(to_json(argval))
        %timeit memoryview(argval)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cache import *  # NOQA
        >>> import utool as ut
        >>> use_hash = None
        >>> func = consensed_cfgstr
        >>> args = ('a', 'b', 'c', 'd')
        >>> kwargs = {}
        >>> key_argx = [0, 1, 2]
        >>> key_kwds = []
        >>> kwdefaults = ut.util_inspect.get_kwdefaults(func)
        >>> argnames   = ut.util_inspect.get_argnames(func)
        >>> get_cfgstr_from_args(func, args, kwargs, key_argx, key_kwds, kwdefaults, argnames)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cache import *  # NOQA
        >>> import utool as ut
        >>> self = ut.LazyList
        >>> use_hash = None
        >>> func = self.append
        >>> args = ('a', 'b')
        >>> kwargs = {}
        >>> key_argx = [1]
        >>> key_kwds = []
        >>> kwdefaults = ut.util_inspect.get_kwdefaults(func)
        >>> argnames   = ut.util_inspect.get_argnames(func)
        >>> get_cfgstr_from_args(func, args, kwargs, key_argx, key_kwds, kwdefaults, argnames)
    """
    #try:
    #fmt_str = '%s(%s)'
    import utool as ut
    hashstr_ = util_hash.hashstr27
    if key_argx is None:
        key_argx = list(range(len(args)))
    if key_kwds is None:
        key_kwds = ut.unique_ordered(list(kwdefaults.keys()) + list(kwargs.keys()))

    #def kwdval(key):
    #    return kwargs.get(key, kwdefaults.get(key, None))
    given_kwargs = ut.merge_dicts(kwdefaults, kwargs)

    arg_hashfmtstr = [argnames[argx] + '=(%s)' for argx in key_argx]
    #kw_hashfmtstr = [kwdefaults.get(key, '???') + '(%s)' for key in key_kwds]
    kw_hashfmtstr = [key + '=(%s)' for key in key_kwds]
    cfgstr_fmt = '_'.join(chain(arg_hashfmtstr, kw_hashfmtstr))
    #print('cfgstr_fmt = %r' % cfgstr_fmt)
    argrepr_iter = (cachestr_repr(args[argx]) for argx in key_argx)
    kwdrepr_iter = (cachestr_repr(given_kwargs[key]) for key in key_kwds)
    if use_hash is None:
        #print('conditional hashing args')
        argcfg_list = [hashstr_(argrepr) if len(argrepr) > 16 else argrepr
                       for argrepr in argrepr_iter]
        kwdcfg_list =  [hashstr_(kwdrepr) if len(kwdrepr) > 16 else kwdrepr
                        for kwdrepr in kwdrepr_iter]
    elif use_hash is True:
        #print('hashing args')
        argcfg_list = [hashstr_(argrepr) for argrepr in argrepr_iter]
        kwdcfg_list =  [hashstr_(kwdrepr) for kwdrepr in kwdrepr_iter]
    else:
        argcfg_list = list(argrepr_iter)
        kwdcfg_list = list(kwdrepr_iter)
    #print('formating args and kwargs')
    cfgstr = cfgstr_fmt % tuple(chain(argcfg_list, kwdcfg_list))
    #print('made cfgstr = %r' % cfgstr)
    return cfgstr


def cached_func(fname=None, cache_dir='default', appname='utool', key_argx=None,
                key_kwds=None, use_cache=None, verbose=None):
    r"""
    Wraps a function with a Cacher object

    uses a hash of arguments as input

    Args:
        fname (str):  file name (defaults to function name)
        cache_dir (unicode): (default = u'default')
        appname (unicode): (default = u'utool')
        key_argx (None): (default = None)
        key_kwds (None): (default = None)
        use_cache (bool):  turns on disk based caching(default = None)

    CommandLine:
        python -m utool.util_cache --exec-cached_func

    Example:
        >>> # ENABLE_DOCTEST
        >>> import utool as ut
        >>> def costly_func(a, b, c='d', *args, **kwargs):
        ...     return ([a] * b, c, args, kwargs)
        >>> ans0 = costly_func(41, 3)
        >>> ans1 = costly_func(42, 3)
        >>> closure_ = ut.cached_func('costly_func', appname='utool_test',
        >>>                           key_argx=[0, 1])
        >>> efficient_func = closure_(costly_func)
        >>> ans2 = efficient_func(42, 3)
        >>> ans3 = efficient_func(42, 3)
        >>> ans4 = efficient_func(41, 3)
        >>> ans5 = efficient_func(41, 3)
        >>> assert ans1 == ans2
        >>> assert ans2 == ans3
        >>> assert ans5 == ans4
        >>> assert ans5 == ans0
        >>> assert ans1 != ans0
    """
    if verbose is None:
        verbose = VERBOSE_CACHE
    def cached_closure(func):
        # from utool import util_decor
        import utool as ut
        fname_ = util_inspect.get_funcname(func) if fname is None else fname
        kwdefaults = util_inspect.get_kwdefaults(func)
        argnames   = util_inspect.get_argnames(func)
        if ut.is_method(func):
            # ignore self for methods
            argnames = argnames[1:]
        cacher = Cacher(fname_, cache_dir=cache_dir, appname=appname,
                        verbose=verbose)
        if use_cache is None:
            use_cache_ = not util_arg.get_argflag('--nocache-' + fname_)
        else:
            use_cache_ = use_cache
        #_dbgdict = dict(fname_=fname_, key_kwds=key_kwds, appname=appname,
        #                key_argx=key_argx, use_cache_=use_cache_)
        #@functools.wraps(func)
        def cached_wraper(*args, **kwargs):
            """
            Cached Wrapper Function

            Additional Kwargs:
                use_cache (bool) : enables cache
            """
            try:
                if verbose > 2:
                    print('[util_cache] computing cached function fname_=%s' %
                          ( fname_,))
                # Implicitly adds use_cache to kwargs
                cfgstr = get_cfgstr_from_args(func, args, kwargs, key_argx,
                                              key_kwds, kwdefaults, argnames)
                if util_cplat.WIN32:
                    # remove potentially invalid chars
                    cfgstr = '_' + util_hash.hashstr27(cfgstr)
                assert cfgstr is not None, 'cfgstr=%r cannot be None' % (cfgstr,)
                use_cache__ = kwargs.pop('use_cache', use_cache_)
                if use_cache__:
                    # Make cfgstr from specified input
                    data = cacher.tryload(cfgstr)
                    if data is not None:
                        return data
                # Cached missed compute function
                data = func(*args, **kwargs)
                # Cache save
                #if use_cache__:
                # TODO: save_cache
                cacher.save(data, cfgstr)
                return data
            #except ValueError as ex:
            # handle protocal error
            except Exception as ex:
                from utool import util_dbg
                _dbgdict2 = dict(key_argx=key_argx, lenargs=len(args),
                                 lenkw=len(kwargs),)
                msg = '\n'.join([
                    '+--- UTOOL --- ERROR IN CACHED FUNCTION',
                    #'dbgdict = ' + utool.repr4(_dbgdict),
                    'dbgdict2 = ' + util_str.repr4(_dbgdict2),
                ])
                util_dbg.printex(ex, msg)
                raise
        # Give function a handle to the cacher object
        cached_wraper = util_decor.preserve_sig(cached_wraper, func)
        cached_wraper.cacher = cacher
        return cached_wraper
    return cached_closure


# --- Global Cache ---

def view_global_cache_dir(appname='default'):
    import utool
    dir_ = utool.get_global_cache_dir(appname=appname)
    utool.view_directory(dir_)


def get_global_cache_dir(appname='default', ensure=False):
    """ Returns (usually) writable directory for an application cache """
    if appname is None or  appname == 'default':
        appname = get_default_appname()
    global_cache_dir = util_cplat.get_app_resource_dir(appname,
                                                       meta_util_constants.global_cache_dname)
    if ensure:
        util_path.ensuredir(global_cache_dir)
    return global_cache_dir


def get_global_shelf_fpath(appname='default', ensure=False):
    """ Returns the filepath to the global shelf """
    global_cache_dir = get_global_cache_dir(appname, ensure=ensure)
    shelf_fpath = join(global_cache_dir, meta_util_constants.global_cache_fname)
    return shelf_fpath


def shelf_open(fpath):
    """
    allows for shelf to be used in with statements

    References:
        http://stackoverflow.com/questions/7489732/easiest-way-to-add-a-function-to-existing-class

    CommandLine:
        python -m utool.util_cache --test-shelf_open

    Example:
        >>> # DISABLE_DOCTEST
        >>> # UNSTABLE_DOCTEST
        >>> import utool as ut
        >>> fpath = ut.unixjoin(ut.ensure_app_resource_dir('utool'), 'testshelf.shelf')
        >>> with ut.shelf_open(fpath) as dict_:
        ...     print(ut.repr4(dict_))
    """
    return contextlib.closing(shelve.open(fpath))


#class YAWShelf(object):
#    def __init__(self, shelf_fpath):
#        self.shelf_fpath = shelf_fpath
#        import shelve
#        self.shelf = shelve.open(shelf_fpath)


class GlobalShelfContext(object):
    """ older class. might need update """
    def __init__(self, appname):
        self.appname = appname

    def __enter__(self):
        #self.shelf = get_global_shelf(self.appname)

        try:
            import dbm
            DBMError = dbm.error
        except Exception:
            DBMError = OSError

        try:
            shelf_fpath = get_global_shelf_fpath(self.appname, ensure=True)
            if VERBOSE:
                print('[cache] open: ' + shelf_fpath)
            self.shelf = shelve.open(shelf_fpath)
        except DBMError as ex:
            from utool import util_dbg
            util_dbg.printex(ex, 'Failed opening shelf_fpath due to bad version, remove and retry',
                             key_list=['shelf_fpath'])
            import utool as ut
            ut.delete(shelf_fpath)
            self.shelf = shelve.open(shelf_fpath)
        except Exception as ex:
            from utool import util_dbg
            util_dbg.printex(ex, 'Failed opening shelf_fpath',
                             key_list=['shelf_fpath'])
            raise
        return self.shelf

    def __exit__(self, type_, value, trace):
        self.shelf.close()
        if trace is not None:
            print('[cache] Error under GlobalShelfContext!: ' + str(value))
            return False  # return a falsey value on error
        #close_global_shelf(self.appname)


def global_cache_read(key, appname='default', **kwargs):
    with GlobalShelfContext(appname) as shelf:
        if 'default' in kwargs:
            return shelf.get(key, kwargs['default'])
        else:
            return shelf[key]


def global_cache_dump(appname='default'):
    shelf_fpath = get_global_shelf_fpath(appname)
    print('shelf_fpath = %r' % shelf_fpath)
    with GlobalShelfContext(appname) as shelf:
        print(util_str.repr4(shelf))


def global_cache_write(key, val, appname='default'):
    """ Writes cache files to a safe place in each operating system """
    with GlobalShelfContext(appname) as shelf:
        shelf[key] = val


def delete_global_cache(appname='default'):
    """ Reads cache files to a safe place in each operating system """
    #close_global_shelf(appname)
    shelf_fpath = get_global_shelf_fpath(appname)
    util_path.remove_file(shelf_fpath, verbose=True, dryrun=False)

#import abc  # abstract base class
#import six


#@six.add_metaclass(abc.ABCMeta)
class Cachable(object):
    """
    Abstract base class.

    This class which enables easy caching of object dictionarys

    must implement get_cfgstr()

    """
    ext = '.cPkl'  # TODO: Capt'n Proto backend to replace pickle backend

    #@abc.abstractmethod
    def get_cfgstr(self):
        return getattr(self, 'cfgstr', 'DEFAULT')
        # return 'DEFAULT'
        # raise NotImplementedError('abstract method')

    #@abc.abstractmethod
    def get_prefix(self):
        # import utool as ut
        return self.__class__.__name__ + '_'
        # return ut.get_funcname(self.__class__) + '_'
        # raise NotImplementedError('abstract method')

    def get_cachedir(self, cachedir=None):
        if cachedir is None:
            if hasattr(self, 'cachedir'):
                cachedir = self.cachedir
            else:
                cachedir = '.'
        return cachedir

    def get_fname(self, cfgstr=None, ext=None):
        # convinience
        return basename(self.get_fpath('', cfgstr=cfgstr, ext=ext))

    def get_fpath(self, cachedir=None, cfgstr=None, ext=None):
        """
        Ignore:
            fname = _fname
            cfgstr = _cfgstr
        """
        _dpath = self.get_cachedir(cachedir)
        _fname = self.get_prefix()
        _cfgstr = self.get_cfgstr() if cfgstr is None else cfgstr
        _ext =   self.ext if ext is None else ext
        fpath = _args2_fpath(_dpath, _fname, _cfgstr, _ext)
        return fpath

    def delete(self, cachedir=None, cfgstr=None, verbose=True or VERBOSE or util_arg.VERBOSE):
        """
        saves query result to directory
        """
        fpath = self.get_fpath(cachedir, cfgstr=cfgstr)
        if verbose:
            print('[Cachable] cache delete: %r' % (basename(fpath),))
        os.remove(fpath)

    @profile
    def save(self, cachedir=None, cfgstr=None, verbose=VERBOSE, quiet=QUIET,
             ignore_keys=None):
        """
        saves query result to directory
        """
        fpath = self.get_fpath(cachedir, cfgstr=cfgstr)
        if verbose:
            print('[Cachable] cache save: %r' % (basename(fpath),))

        if hasattr(self, '__getstate__'):
            statedict = self.__getstate__()
        else:
            statedict = self.__dict__

        if ignore_keys is None:
            save_dict = statedict
        else:
            save_dict = {key: val
                         for (key, val) in six.iteritems(statedict)
                         if key not in ignore_keys}

        util_io.save_data(fpath, save_dict)
        return fpath
        #save_cache(cachedir, '', cfgstr, self.__dict__)
        #with open(fpath, 'wb') as file_:
        #    pickle.dump(self.__dict__, file_)

    def _unsafe_load(self, fpath, ignore_keys=None):
        loaded_dict = util_io.load_data(fpath)
        if ignore_keys is not None:
            for key in ignore_keys:
                if key in loaded_dict:
                    del loaded_dict[key]
        if hasattr(self, '__setstate__'):
            self.__setstate__(loaded_dict)
        else:
            self.__dict__.update(loaded_dict)
        #with open(fpath, 'rb') as file_:
        #    loaded_dict = pickle.load(file_)
        #    self.__dict__.update(loaded_dict)

    def glob_valid_targets(self, cachedir=None, partial_cfgstr=''):
        from utool import util_path
        prefix = self.get_prefix()
        pattern = prefix + '*' + partial_cfgstr + '*' + self.ext
        cachedir = self.get_cachedir(cachedir)
        valid_targets = util_path.glob(cachedir, pattern, recursive=False)
        return valid_targets

    def fuzzyload(self, cachedir=None, partial_cfgstr='', **kwargs):
        """
        Try and load from a partially specified configuration string
        """
        valid_targets = self.glob_valid_targets(cachedir, partial_cfgstr)
        if len(valid_targets) != 1:
            import utool as ut
            msg = 'need to further specify target. valid_targets=%s' % (ut.repr3(valid_targets,))
            raise ValueError(msg)
        fpath = valid_targets[0]
        self.load(fpath=fpath, **kwargs)

    @profile
    def load(self, cachedir=None, cfgstr=None, fpath=None, verbose=None,
             quiet=QUIET, ignore_keys=None):
        """
        Loads the result from the given database
        """
        if verbose is None:
            verbose = getattr(self, 'verbose', VERBOSE)
        if fpath is None:
            fpath = self.get_fpath(cachedir, cfgstr=cfgstr)
        if verbose:
            print('[Cachable] cache tryload: %r' % (basename(fpath),))
        try:
            self._unsafe_load(fpath, ignore_keys)
            if verbose:
                print('... self cache hit: %r' % (basename(fpath),))
        except ValueError as ex:
            import utool as ut
            msg = '[!Cachable] Cachable(%s) is likely corrupt' % (self.get_cfgstr())
            print('CORRUPT fpath = %s' % (fpath,))
            ut.printex(ex, msg, iswarning=True)
            raise
        #except BadZipFile as ex:
        except zipfile.error as ex:
            import utool as ut
            msg = '[!Cachable] Cachable(%s) has bad zipfile' % (self.get_cfgstr())
            print('CORRUPT fpath = %s' % (fpath,))
            ut.printex(ex, msg, iswarning=True)
            raise
            #if exists(fpath):
            #    #print('[Cachable] Removing corrupted file: %r' % fpath)
            #    #os.remove(fpath)
            #    raise hsexcept.HotsNeedsRecomputeError(msg)
            #else:
            #    raise Exception(msg)
        except IOError as ex:
            import utool as ut
            if not exists(fpath):
                msg = '... self cache miss: %r' % (basename(fpath),)
                if verbose:
                    print(msg)
                raise
            print('CORRUPT fpath = %s' % (fpath,))
            msg = '[!Cachable] Cachable(%s) is corrupt' % (self.get_cfgstr())
            ut.printex(ex, msg, iswarning=True)
            raise
        except Exception as ex:
            import utool as ut
            ut.printex(ex, 'unknown exception while loading query result')
            raise


def get_lru_cache(max_size=5):
    """
    Args:
        max_size (int):

    References:
        https://github.com/amitdev/lru-dict

    CommandLine:
        python -m utool.util_cache --test-get_lru_cache

    Example:
        >>> # DISABLE_DOCTEST
        >>> # UNSTABLE_DOCTEST
        >>> from utool.util_cache import *  # NOQA
        >>> import utool as ut  # NOQA
        >>> max_size = 5
        >>> # execute function
        >>> cache_obj = get_lru_cache(max_size)
        >>> cache_obj[1] = 1
        >>> cache_obj[2] = 2
        >>> cache_obj[3] = 3
        >>> cache_obj[4] = 4
        >>> cache_obj[5] = 5
        >>> cache_obj[6] = 6
        >>> # verify results
        >>> result = ut.repr2(dict(cache_obj), nl=False)
        >>> print(result)
        {2: 2, 3: 3, 4: 4, 5: 5, 6: 6}
    """
    USE_C_LRU = False
    if USE_C_LRU:
        import lru
        cache_obj = lru.LRU(max_size)
    else:
        cache_obj = LRUDict(max_size)
    return cache_obj


class LRUDict(object):
    """
    Pure python implementation for lru cache fallback

    References:
        http://www.kunxi.org/blog/2014/05/lru-cache-in-python/

    Args:
        max_size (int): (default = 5)

    Returns:
        LRUDict: cache_obj

    CommandLine:
        python -m utool.util_cache --test-LRUDict

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_cache import *  # NOQA
        >>> max_size = 5
        >>> self = LRUDict(max_size)
        >>> for count in range(0, 5):
        ...     self[count] = count
        >>> print(self)
        >>> self[0]
        >>> for count in range(5, 8):
        ...     self[count] = count
        >>> print(self)
        >>> del self[5]
        >>> assert 4 in self
        >>> result = ('self = %r' % (self,))
        >>> print(result)
        self = LRUDict({
            4: 4,
            0: 0,
            6: 6,
            7: 7,
        })
    """

    def __init__(self, max_size):
        self._max_size = max_size
        self._cache = collections.OrderedDict()

    def has_key(self, item):
        return item in self

    def __contains__(self, item):
        return item in self._cache

    def __delitem__(self, key):
        del self._cache[key]

    def __str__(self):
        import utool as ut
        return ut.repr4(self._cache, nl=False)

    def __repr__(self):
        import utool as ut
        return 'LRUDict(' + ut.repr4(self._cache) + ')'
        #return repr(self._cache)

    def __iter__(self):
        return iter(self._cache)

    def items(self):
        return self._cache.items()

    def keys(self):
        return self._cache.keys()

    def values(self):
        return self._cache.values()

    def iteritems(self):
        return self._cache.iteritems()

    def iterkeys(self):
        return self._cache.iterkeys()

    def itervalues(self):
        return self._cache.itervalues()

    def clear(self):
        return self._cache.clear()

    def __len__(self):
        return len(self._cache)

    def __getitem__(self, key):
        try:
            value = self._cache.pop(key)
            self._cache[key] = value
            return value
        except KeyError:
            raise

    def __setitem__(self, key, value):
        try:
            self._cache.pop(key)
        except KeyError:
            if len(self._cache) >= self._max_size:
                self._cache.popitem(last=False)
        self._cache[key] = value


def time_different_diskstores():
    """
    %timeit shelf_write_test()    # 15.1 ms per loop
    %timeit cPickle_write_test()  # 1.26 ms per loop

    %timeit shelf_read_test()     # 8.77 ms per loop
    %timeit cPickle_read_test()   # 2.4 ms per loop
    %timeit cPickle_read_test2()  # 2.35 ms per loop

    %timeit json_read_test()
    %timeit json_write_test()
    """
    import utool as ut
    import simplejson as json
    shelf_path = 'test.shelf'
    json_path = 'test.json'
    cpkl_path = 'test.pkl'
    size = 1000
    dict_ = {str(key): str(uuid.uuid4()) for key in range(size)}
    ut.delete(cpkl_path)
    ut.delete(json_path)
    ut.delete(shelf_path)

    def shelf_write_test():
        with ut.shelf_open(shelf_path) as shelf_dict:
            shelf_dict.update(dict_)

    def shelf_read_test():
        with ut.shelf_open(shelf_path) as shelf_dict:
            test = {key: val for key, val in six.iteritems(shelf_dict)}
        assert len(test) > 0

    def json_write_test():
        with open(json_path, 'wb') as outfile:
            json.dump(dict_, outfile)

    def cPickle_write_test():
        with open(cpkl_path, 'wb') as outfile:
            pickle.dump(dict_, outfile)

    def cPickle_read_test():
        with open(cpkl_path, 'rb') as outfile:
            test = {key: val for key, val in six.iteritems(pickle.load(outfile))}
        assert len(test) > 0

    def cPickle_read_test2():
        with open(cpkl_path, 'rb') as outfile:
            test = pickle.load(outfile)
        assert len(test) > 0

    shelf_write_test()
    shelf_read_test()
    #json_write_test()
    #json_read_test()
    cPickle_write_test()
    cPickle_read_test()
    cPickle_read_test2()


class KeyedDefaultDict(util_dict.DictLike):
    def __init__(self, default_func, *args, **kwargs):
        self._default_func = default_func
        self._args = args
        self._kwargs = kwargs
        self._internal = {}

    def setitem(self, key, value):
        self._internal[key] = value

    def getitem(self, key):
        if key not in self._internal:
            value = self._default_func(key, *self._args, **self._kwargs)
            self._internal[key] = value
        return self._internal[key]

    def keys(self):
        return self._internal.keys()

    def values(self):
        return self._internal.values()


# @six.add_metaclass(util_class.ReloadingMetaclass)
@util_class.reloadable_class
class LazyDict(object):
    #class LazyDict(collections.Mapping):
    """
    Hacky dictionary where values that are functions are counted as lazy


    CommandLine:
        python -m utool.util_cache --exec-LazyDict

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_cache import *  # NOQA
        >>> import utool as ut
        >>> self = ut.LazyDict()
        >>> self['foo'] = lambda: 5
        >>> self['bar'] = 4
        >>> try:
        >>>     self['foo'] = lambda: 9
        >>>     assert False, 'should not be able to override computable functions'
        >>> except ValueError:
        >>>     pass
        >>> self['biz'] = lambda: 9
        >>> d = {}
        >>> d.update(**self)
        >>> self['spam'] = lambda: 'eggs'
        >>> self.printinfo()
        >>> print(self.tostring(is_eager=False))
    """
    def __init__(self, other=None, is_eager=True, verbose=False, reprkw=None,
                 mutable=False,
                 **kwargs):
        # Registered lazy evaluations
        self._eval_funcs = {}
        # Computed results
        self._stored_results = {}
        self.infer_lazy_vals_hack = True
        self._is_eager = is_eager
        self._verbose = verbose
        self.reprkw = dict(is_eager=False, nl=False)
        self._mutable = mutable
        if reprkw is not None:
            self.reprkw.update(**reprkw)
        if other is not None:
            self.update(other)
        if len(kwargs) > 0:
            self.update(kwargs)

    # --- direct interface

    def set_lazy_func(self, key, func):
        assert util_type.is_funclike(func), 'func must be a callable'
        #if key in self._stored_results:
        #    raise ValueError(
        #        ('Cannot add new lazy function for key=%r'
        #         'that has been computed') % (key,))
        #if key in self._stored_results:
        if not self._mutable and key in self.reconstructable_keys():
            raise ValueError(
                ('Cannot overwrite lazy function for key=%r') % (key,))
        self._eval_funcs[key] = func

    def setitem(self, key, value):
        # HACK, lazy funcs should all be registered
        # this should should always just set a value
        if not self._mutable and key in self.reconstructable_keys():
            raise ValueError(
                ('Cannot overwrite lazy function for key=%r') % (key,))
        if (self.infer_lazy_vals_hack and
             util_type.is_funclike(value)):
            self.set_lazy_func(key, value)
        else:
            self._stored_results[key] = value

    def getitem(self, key, is_eager=None):
        if is_eager is None:
            is_eager = self._is_eager
        if is_eager:
            return self.eager_eval(key)
        else:
            return self.lazy_eval(key)

    def nocache_eval(self, key):
        """ forces function evaluation """
        func_ = self._eval_funcs[key]
        value = func_()
        return value

    def eager_eval(self, key):
        if key in self._stored_results:
            value  = self._stored_results[key]
        else:
            if self._verbose:
                print('[util_cache] Evaluating key=%r' % (key,))
            value = self.nocache_eval(key)
            self._stored_results[key] = value
        return value

    def lazy_eval(self, key):
        if key in self._stored_results:
            value  = self._stored_results[key]
        else:
            value = self._eval_funcs[key]
        return value

    def clear_evaluated(self):
        for key in list(self.evaluated_keys()):
            del self._stored_results[key]

    def clear_stored(self, keys=None):
        if keys is None:
            keys = list(self.stored_keys())
        for key in keys:
            del self._stored_results[key]

    def stored_keys(self):
        """ keys whose vals that have been explicitly set or evaluated """
        return self._stored_results.keys()

    def reconstructable_keys(self):
        """ only keys whose vals that have been set with a backup func """
        return set(self._eval_funcs.keys())

    def all_keys(self):
        return set(self.stored_keys()).union(set(self.reconstructable_keys()))

    def unevaluated_keys(self):
        """ keys whose vals can be constructed but have not been """
        return set(self.reconstructable_keys()) - set(self.stored_keys())

    def evaluated_keys(self):
        """ only keys whose vals have been evaluated from a stored function """
        return set(self.reconstructable_keys()) - set(self.unevaluated_keys())

    def nonreconstructable_keys(self):
        """ only keys whose vals that have been explicitly set without a backup func """
        return set(self.all_keys()) - self.reconstructable_keys()

    def cached_keys(self):
        """ only keys whose vals that have been explicitly set without a backup func """
        return set(self.nonreconstructable_keys()).union(set(self.evaluated_keys()))

    def printinfo(self):
        print('nonreconstructable_keys = %s' % (self.nonreconstructable_keys(),))
        print('reconstructable_keys = %s' % (self.reconstructable_keys(),))
        print('evaluated_keys = %s' % (self.evaluated_keys(),))
        print('unevaluated_keys = %s' % (self.unevaluated_keys(),))

    def asdict(self, is_eager=None):
        dict_ = {key: self.getitem(key, is_eager) for key in self.keys()}
        return dict_

    def tostring(self, is_eager=None, keys=None, **kwargs):
        import utool as ut
        dict_ = self.asdict(is_eager=is_eager)
        class AwakeFaceRepr(object):
            def __repr__(self):
                return '!'
                # return '(o.o)'
                # return "٩(ˊᗜˋ*)و"
        class SleepFaceRepr(object):
            def __repr__(self):
                return 'z'
                # return '(-_-)'
                # return '(ᵕ≀ᵕ)'
        for key in self.evaluated_keys():
            # dict_[key] = '!'
            dict_[key] = AwakeFaceRepr()
        for key in self.unevaluated_keys():
            # dict_[key] = 'z'
            dict_[key] = SleepFaceRepr()
        if keys is not None:
            dict_ = ut.dict_subset(dict_, keys)
        return ut.repr2(dict_, **kwargs)

    # --- dict interface

    def get(self, key, *d):
        if len(d) > 1:
            raise ValueError('can only specify one default')
        elif len(d) == 1:
            #assert len(d) == 0, 'no support for default yet'
            if key not in self:
                return d[0]
        return self.getitem(key, self._is_eager)

    def update(self, dict_, **kwargs):
        for key, val in six.iteritems(dict_):
            self[key] = val
        for key, val in six.iteritems(kwargs):
            self[key] = val

    def keys(self):
        return self.all_keys()

    def values(self):
        return [self[key] for key in self.keys()]

    def items(self):
        return [(key, self[key]) for key in self.keys()]

    def __setitem__(self, key, value):
        self.setitem(key, value)

    def __getitem__(self, key):
        return self.get(key)

    def __delitem__(self, key):
        if key not in self.keys():
            raise KeyError(key)
        if key in self._eval_funcs:
            del self._eval_funcs[key]
        if key in self._stored_results:
            del self._stored_results[key]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    def __str__(self):
        return self.tostring()

    def __repr__(self):
        return self.tostring(**self.reprkw)

    #def __getstate__(self):
    #    state_dict = self.asdict()
    #    return state_dict

    #def __setstate__(self, state_dict):
    #    self._stored_results.update(state_dict)


@six.add_metaclass(util_class.ReloadingMetaclass)
class LazyList(object):
    """ very hacky list implemented as a dictionary """
    def __init__(self, **kwargs):
        self._hackstore = LazyDict(**kwargs)

    def __len__(self):
        return len(self._hackstore)

    def __getitem__(self, index):
        try:
            return self._hackstore[index]
        except KeyError:
            #raise ValueError('index=%r out of bounds' % (index,))
            raise ValueError(
                'index=%r out of bounds or error computing lazy value.' % (
                    index,))

    def append(self, item):
        self._hackstore[len(self._hackstore)] = item

    def tolist(self):
        return self._hackstore.values()

    def __iter__(self):
        for index in range(len(self)):
            yield self[index]


if __name__ == '__main__':
    """
    CommandLine:
        python -c "import utool, utool.util_cache; utool.doctest_funcs(utool.util_cache)"
        python -m utool.util_cache
        python -m utool.util_cache --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import os
from os.path import join, normpath, pathsep, dirname  # NOQA


def get_regstr(regtype, var, val):
    regtype_map = {
        'REG_EXPAND_SZ': 'hex(2):',
        'REG_DWORD': 'dword:',
        'REG_BINARY': None,
        'REG_MULTI_SZ': None,
        'REG_SZ': '',
    }
    # It is not a good idea to write these variables...
    EXCLUDE = ['USERPROFILE', 'USERNAME', 'SYSTEM32']
    if var in EXCLUDE:
        return ''
    def quotes(str_):
        return '"' + str_.replace('"', r'\"') + '"'
    sanitized_var = quotes(var)
    if regtype == 'REG_EXPAND_SZ':
        # Weird encoding
        #bin_ = binascii.hexlify(hex_)
        #val_ = ','.join([''.join(hex2) for hex2 in hex2zip])
        #import binascii  # NOQA
        x = val
        ascii_ = x.encode("ascii")
        hex_ = ascii_.encode("hex")
        hex_ = x.encode("hex")
        hex2zip = zip(hex_[0::2], hex_[1::2])
        spacezip = [('0', '0')] * len(hex2zip)
        hex3zip = zip(hex2zip, spacezip)
        sanitized_val = ','.join([''.join(hex2) + ',' + ''.join(space) for hex2, space in hex3zip])
    elif regtype == 'REG_DWORD':
        sanitized_val = '%08d' % int(val)
    else:
        sanitized_val = quotes(val)
    # Comment with the human-readable nonhex version of the string
    comment = '; ' + var + '=' + val
    regstr = sanitized_var + '=' + regtype_map[regtype] + sanitized_val
    return comment + '\n' + regstr


def make_regfile_str(key, varval_list, rtype):
    # Input: list of (var, val) tuples
    # key to put varval list in
    # rtype - type of registry variables
    envtxt_list = ['Windows Registry Editor Version 5.00',
                   '',
                   key]
    print('\n'.join(map(repr, varval_list)))
    varval_list = filter(lambda x: isinstance(x, tuple), varval_list)
    vartxt_list = [get_regstr(rtype, var, val) for (var, val) in varval_list]
    envtxt_list.extend(vartxt_list)
    regfile_str = '\n'.join(envtxt_list)
    return regfile_str


def add_to_win32_PATH(script_fpath, *add_path_list):
    r"""
    Writes a registery script to update the PATH variable into the sync registry

    CommandLine:
        python -m utool.util_win32 --test-add_to_win32_PATH --newpath "C:\Program Files (x86)\Graphviz2.38\bin"

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_win32 import *  # NOQA
        >>> script_fpath = join(ut.truepath('~'), 'Sync/win7/registry', 'UPDATE_PATH.reg')
        >>> new_path = ut.get_argval('--newpath', str, default=None)
        >>> result = add_to_win32_PATH(script_fpath, new_path)
        >>> print(result)
    """
    import utool as ut
    write_dir = dirname(script_fpath)
    key = r'[HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Session Manager\Environment]'
    rtype = 'REG_EXPAND_SZ'
    # Read current PATH values
    win_pathlist = list(os.environ['PATH'].split(os.path.pathsep))
    new_path_list = ut.unique_ordered(win_pathlist + list(add_path_list))
    #new_path_list = unique_ordered(win_pathlist, rob_pathlist)
    print('\n'.join(new_path_list))
    pathtxt = pathsep.join(new_path_list)
    varval_list = [('Path', pathtxt)]
    regfile_str = make_regfile_str(key, varval_list, rtype)
    ut.view_directory(write_dir)
    print(regfile_str)
    ut.writeto(script_fpath, regfile_str, mode='wb')
    print('Please have an admin run the script. You may need to restart')


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_win32
        python -m utool.util_win32 --allexamples
        python -m utool.util_win32 --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
In this module:
    * a metaclass allowing for reloading of single class instances
    * functions to autoinject methods into a class upon instance creation.
    * A wrapper class allowing an object's properties to be used as kwargs
    * a metaclass to forward properties to another class

    ReloadingMetaclass
    KwargsWrapper
"""
from __future__ import absolute_import, division, print_function
import sys
import six
import types
import functools
import operator as op
try:
    from collections.abc import Mapping
except Exception:
    from collections import Mapping
from collections import defaultdict
from utool import util_inject
from utool import util_set
from utool import util_arg
from utool._internal.meta_util_six import get_funcname, get_funcglobals
print, rrr, profile = util_inject.inject2(__name__)


# Registers which classes have which attributes
# FIXME: this might cause memory leaks
# FIXME: this does cause weird reimport behavior
__CLASSTYPE_ATTRIBUTES__ = defaultdict(util_set.oset)
__CLASSTYPE_POSTINJECT_FUNCS__ = defaultdict(util_set.oset)
__CLASSNAME_CLASSKEY_REGISTER__ = defaultdict(util_set.oset)


#_rrr = rrr
#def rrr(verbose=True):
#    """ keep registered functions through reloads ? """
#    global __CLASSTYPE_ATTRIBUTES__
#    global __CLASSTYPE_POSTINJECT_FUNCS__
#    cta = __CLASSTYPE_ATTRIBUTES__.copy()
#    ctpif = __CLASSTYPE_POSTINJECT_FUNCS__.copy()
#    rrr_(verbose=verbose)
#    __CLASSTYPE_ATTRIBUTES__ = cta
#    __CLASSTYPE_POSTINJECT_FUNCS__ = ctpif


QUIET_CLASS = util_arg.get_argflag(('--quiet-class', '--quietclass'))
VERBOSE_CLASS = (
    util_arg.get_argflag(('--verbose-class', '--verbclass')) or
    (not QUIET_CLASS and util_arg.VERYVERBOSE))


def inject_instance(self, classkey=None, allow_override=False,
                    verbose=VERBOSE_CLASS, strict=True):
    """
    Injects an instance (self) of type (classkey)
    with all functions registered to (classkey)

    call this in the __init__ class function

    Args:
        self: the class instance
        classkey: key for a class, preferably the class type itself, but it
            doesnt have to be

    SeeAlso:
        make_class_method_decorator

    Example:
        >>> # DISABLE_DOCTEST
        >>> # DOCTEST_DISABLE
        >>> utool.make_class_method_decorator(InvertedIndex)(smk_debug.invindex_dbgstr)
        >>> utool.inject_instance(invindex)
    """
    import utool as ut
    if verbose:
        print('[util_class] begin inject_instance')
    try:
        if classkey is None:
            # Probably should depricate this block of code
            # It tries to do too much
            classkey = self.__class__
            if classkey == 'ibeis.gui.models_and_views.IBEISTableView':
                # HACK HACK HACK
                try:
                    from guitool_ibeis.__PYQT__ import QtWidgets  # NOQA
                except ImportError:
                    from guitool.__PYQT__ import QtWidgets  # NOQA
                classkey = QtWidgets.QAbstractItemView
            if len(__CLASSTYPE_ATTRIBUTES__[classkey]) == 0:
                print('[utool] Warning: no classes of type %r are registered' % (classkey,))
                print('[utool] type(self)=%r, self=%r' % (type(self), self)),
                print('[utool] Checking to see if anybody else was registered...')
                print('[utool] __CLASSTYPE_ATTRIBUTES__ = ' +
                      ut.repr4(__CLASSTYPE_ATTRIBUTES__.keys()))
                for classtype_, _ in six.iteritems(__CLASSTYPE_ATTRIBUTES__):
                    isinstance(self, classtype_)
                    classkey = classtype_
                    print('[utool] Warning: using subclass=%r' % (classtype_,))
                    break
        func_list = __CLASSTYPE_ATTRIBUTES__[classkey]
        if verbose:
            print('[util_class] injecting %d methods\n   with classkey=%r\n   into %r'
                  % (len(func_list), classkey, self,))
        for func in func_list:
            if VERBOSE_CLASS:
                print('[util_class] * injecting %r' % (func,))
            method_name = None
            # Allow user to register tuples for aliases
            if isinstance(func, tuple):
                func, method_name = func
            inject_func_as_method(self, func, method_name=method_name,
                                  allow_override=allow_override, verbose=verbose)
    except Exception as ex:
        ut.printex(ex, 'ISSUE WHEN INJECTING %r' % (classkey,),
                      iswarning=not strict)
        if strict:
            raise


def postinject_instance(self, classkey, verbose=VERBOSE_CLASS):
    if verbose:
        print('[util_class] Running postinject functions on %r' % (self,))
    for func in __CLASSTYPE_POSTINJECT_FUNCS__[classkey]:
        func(self)
    if verbose:
        print('[util_class] Finished injecting instance self=%r' % (self,))


def inject_all_external_modules(self, classname=None,
                                allow_override='override+warn',
                                strict=True):
    """
    dynamically injects registered module methods into a class instance

    FIXME: naming convention and use this in all places where this clas is used
    """
    #import utool as ut
    if classname is None:
        classname = self.__class__.__name__
    #import utool as ut
    #ut.embed()

    NEW = True
    if NEW:
        classkey_list = [key for key in __CLASSTYPE_ATTRIBUTES__
                         if key[0] == classname]
    else:
        injected_modules = get_injected_modules(classname)
        # the variable must be named CLASS_INJECT_KEY
        # and only one class can be specified per module.
        classkey_list = [module.CLASS_INJECT_KEY
                         for module in injected_modules]

    for classkey in classkey_list:
        inject_instance(
            self, classkey=classkey,
            allow_override=allow_override, strict=False)

    for classkey in classkey_list:
        postinject_instance(
            self, classkey=classkey)


def reload_injected_modules(classname):
    injected_modules = get_injected_modules(classname)
    for module in injected_modules:
        if hasattr(module, 'rrr'):
            module.rrr()
        else:
            import imp
            print('rrr not defined in module=%r' % (module,))
            imp.reload(module)


def get_injected_modules(classname):
    r"""
    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_class import __CLASSNAME_CLASSKEY_REGISTER__  # NOQA
    """
    modname_list = __CLASSNAME_CLASSKEY_REGISTER__[classname]

    injected_modules = []
    for modstr in modname_list:
        parts = modstr.split('.')
        pkgname = '.'.join(parts[:-1])
        modname = parts[-1]
        try:
            exec('from %s import %s' % (pkgname, modname, ), globals(), locals())
            module = eval(modname)
            injected_modules.append(module)
        except ImportError as ex:
            ut.printex(ex, 'Cannot load package=%r, module=%r' % (pkgname, modname, ))
    return injected_modules


def autogen_import_list(classname, conditional_imports=None):
    import utool as ut
    #ut.embed()
    #line_list = []
    line_list = ['import sys  # NOQA']
    for modname in __CLASSNAME_CLASSKEY_REGISTER__[classname]:
        # <super hacky>
        condition = None
        for x in conditional_imports:
            if modname == x[1]:
                condition = x[0]
        # </super hacky>
        parts = modname.split('.')
        frompart = '.'.join(parts[:-1])
        imppart = parts[-1]
        #line = 'from %s import %s  # NOQA' % (frompart, imppart)
        if condition is None:
            line = 'from %s import %s' % (frompart, imppart)
        else:
            line = ut.codeblock(
                '''
                if not ut.get_argflag({condition}) or '{frompart}' in sys.modules:
                    from {frompart} import {imppart}
                ''').format(condition=condition, frompart=frompart,
                            imppart=imppart)
        line_list.append(line)
    src = '\n'.join(line_list)
    return src


def autogen_explicit_injectable_metaclass(classname, regen_command=None,
                                          conditional_imports=None):
    r"""
    Args:
        classname (?):

    Returns:
        ?:

    CommandLine:
        python -m utool.util_class --exec-autogen_explicit_injectable_metaclass

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_class import *  # NOQA
        >>> from utool.util_class import  __CLASSTYPE_ATTRIBUTES__  # NOQA
        >>> import ibeis
        >>> import ibeis.control.IBEISControl
        >>> classname = ibeis.control.controller_inject.CONTROLLER_CLASSNAME
        >>> result = autogen_explicit_injectable_metaclass(classname)
        >>> print(result)
    """
    import utool as ut
    vals_list = []

    def make_redirect(func):
        # PRESERVES ALL SIGNATURES WITH EXECS
        src_fmt = r'''
        def {funcname}{defsig}:
            """ {orig_docstr}"""
            return {orig_funcname}{callsig}
        '''
        from utool._internal import meta_util_six
        orig_docstr = meta_util_six.get_funcdoc(func)
        funcname = meta_util_six.get_funcname(func)
        orig_funcname = modname.split('.')[-1] + '.' + funcname
        orig_docstr = '' if orig_docstr is None else orig_docstr
        import textwrap
        # Put wrapped function into a scope
        import inspect
        argspec = inspect.getargspec(func)
        (args, varargs, varkw, defaults) = argspec
        defsig = inspect.formatargspec(*argspec)
        callsig = inspect.formatargspec(*argspec[0:3])
        src_fmtdict = dict(funcname=funcname, orig_funcname=orig_funcname,
                           defsig=defsig, callsig=callsig,
                           orig_docstr=orig_docstr)
        src = textwrap.dedent(src_fmt).format(**src_fmtdict)
        return src

    src_list = []

    for classkey, vals in __CLASSTYPE_ATTRIBUTES__.items():
        modname = classkey[1]
        if classkey[0] == classname:
            vals_list.append(vals)
            for func in vals:
                src = make_redirect(func)
                src = ut.indent(src)
                src = '\n'.join([_.rstrip() for _ in src.split('\n')])
                src_list.append(src)

    if regen_command is None:
        regen_command = 'FIXME None given'

    module_header = ut.codeblock(
        """
        # -*- coding: utf-8 -*-
        """ + ut.TRIPLE_DOUBLE_QUOTE + """
        Static file containing autogenerated functions for {classname}
        Autogenerated on {autogen_time}

        RegenCommand:
            {regen_command}
        """ + ut.TRIPLE_DOUBLE_QUOTE + """

        from __future__ import absolute_import, division, print_function
        import utool as ut

        """).format(
            autogen_time=ut.get_timestamp(),
            regen_command=regen_command,
            classname=classname)

    depends_module_block = autogen_import_list(classname, conditional_imports)
    inject_statement_fmt = ("print, rrr, profile = "
                            "ut.inject2(__name__, '[autogen_explicit_inject_{classname}]')")
    inject_statement = inject_statement_fmt.format(classname=classname)

    source_block_lines = [
        module_header,
        depends_module_block,
        inject_statement,
        '\n',
        'class ExplicitInject' + classname + '(object):',
    ] + src_list
    source_block = '\n'.join(source_block_lines)

    source_block = ut.autoformat_pep8(source_block, aggressive=2)
    return source_block


def make_class_method_decorator(classkey, modname=None):
    """
    register a class to be injectable
    classkey is a key that identifies the injected class
    REMEMBER to call inject_instance in __init__

    Args:
        classkey : the class to be injected into
        modname : the global __name__ of the module youa re injecting from

    Returns:
        closure_decorate_class_method (func): decorator for injectable methods

    Example:
        >>> # ENABLE_DOCTEST
        >>> import utool as ut
        >>> class CheeseShop(object):
        ...     def __init__(self):
        ...         import utool as ut
        ...         ut.inject_all_external_modules(self)
        >>> cheeseshop_method = ut.make_class_method_decorator(CheeseShop)
        >>> shop1 = CheeseShop()
        >>> assert not hasattr(shop1, 'has_cheese'), 'have not injected yet'
        >>> @cheeseshop_method
        >>> def has_cheese(self):
        >>>     return False
        >>> shop2 = CheeseShop()
        >>> assert shop2.has_cheese() is False, 'external method not injected'
        >>> print('Cheese shop does not have cheese. All is well.')
    """
    global __APP_MODNAME_REGISTER__
    #if util_arg.VERBOSE or VERBOSE_CLASS:
    if VERBOSE_CLASS:
        print('[util_class] register via make_class_method_decorator classkey=%r, modname=%r'
              % (classkey, modname))
    if modname == '__main__':
        # skips reinjects into main
        print('WARNING: cannot register classkey=%r functions as __main__' % (classkey,))
        return lambda func: func
    # register that this module was injected into
    if isinstance(classkey, tuple):
        classname, _ = classkey
        __CLASSNAME_CLASSKEY_REGISTER__[classname].append(modname)
    elif isinstance(classkey, type):
        classname = classkey.__name__
        if modname is not None:
            assert modname == classkey.__module__, (
                'modname=%r does not agree with __module__=%r' % (
                    modname, classkey.__module__))
        modname = classkey.__module__
        # Convert to new classkey format
        classkey = (classname, modname)
        __CLASSNAME_CLASSKEY_REGISTER__[classname].append(modname)
    else:
        print('Warning not using classkey for %r %r' % (classkey, modname))
        raise AssertionError('classkey no longer supported. Use class_inject_key instead')
    closure_decorate_class_method = functools.partial(decorate_class_method,
                                                      classkey=classkey)
    return closure_decorate_class_method


def make_class_postinject_decorator(classkey, modname=None):
    """
    Args:
        classkey : the class to be injected into
        modname : the global __name__ of the module youa re injecting from

    Returns:
        closure_decorate_postinject (func): decorator for injectable methods

    SeeAlso:
        make_class_method_decorator
    """
    if util_arg.VERBOSE or VERBOSE_CLASS:
        print('[util_class] register class_postinject classkey=%r, modname=%r'
              % (classkey, modname))
    if modname == '__main__':
        print('WARNING: cannot register class functions as __main__')
        # skips reinjects into main
        return lambda func: func
    closure_decorate_postinject = functools.partial(decorate_postinject,
                                                    classkey=classkey)
    return closure_decorate_postinject


def decorate_class_method(func, classkey=None, skipmain=False):
    """
    Will inject all decorated function as methods of classkey

    classkey is some identifying string, tuple, or object

    func can also be a tuple
    """
    #import utool as ut
    global __CLASSTYPE_ATTRIBUTES__
    assert classkey is not None, 'must specify classkey'
    #if not (skipmain and ut.get_caller_modname() == '__main__'):
    __CLASSTYPE_ATTRIBUTES__[classkey].append(func)
    return func


def decorate_postinject(func, classkey=None, skipmain=False):
    """
    Will perform func with argument self after inject_instance is called on classkey

    classkey is some identifying string, tuple, or object
    """
    #import utool as ut
    global __CLASSTYPE_POSTINJECT_FUNCS__
    assert classkey is not None, 'must specify classkey'
    #if not (skipmain and ut.get_caller_modname() == '__main__'):
    __CLASSTYPE_POSTINJECT_FUNCS__[classkey].append(func)
    return func


def get_method_func(method):
    try:
        return method.im_func if six.PY2 else method.__func__
    except AttributeError:
        # check if this is a method-wrapper type
        if isinstance(method, type(all.__call__)):
            # in which case there is no underlying function
            return None
        raise


def inject_func_as_method(self, func, method_name=None, class_=None,
                          allow_override=False, allow_main=False,
                          verbose=True, override=None, force=False):
    """ Injects a function into an object as a method

    Wraps func as a bound method of self. Then injects func into self
    It is preferable to use make_class_method_decorator and inject_instance

    Args:
       self (object): class instance
       func : some function whos first arugment is a class instance
       method_name (str) : default=func.__name__, if specified renames the method
       class_ (type) : if func is an unbound method of this class


    References:
        http://stackoverflow.com/questions/1015307/python-bind-an-unbound-method
    """
    if override is not None:
        # TODO depcirate allow_override
        allow_override = override
    if method_name is None:
        method_name = get_funcname(func)
    if force:
        allow_override = True
        allow_main = True
    old_method = getattr(self, method_name, None)
    # Bind function to the class instance
    #new_method = types.MethodType(func, self, self.__class__)
    new_method = func.__get__(self, self.__class__)
    #new_method = profile(func.__get__(self, self.__class__))

    if old_method is not None:
        old_im_func = get_method_func(old_method)
        new_im_func = get_method_func(new_method)
        if not allow_main and old_im_func is not None and (
                get_funcglobals(old_im_func)['__name__'] != '__main__' and
                get_funcglobals(new_im_func)['__name__'] == '__main__'):
            if True or VERBOSE_CLASS:
                print('[util_class] skipping re-inject of %r from __main__' % method_name)
            return
        if old_method is new_method or old_im_func is new_im_func:
            #if verbose and util_arg.NOT_QUIET:
            #    print('WARNING: Skipping injecting the same function twice: %r' % new_method)
                #print('WARNING: Injecting the same function twice: %r' % new_method)
            return
        elif allow_override is False:
            raise AssertionError(
                'Overrides are not allowed. Already have method_name=%r' %
                (method_name))
        elif allow_override == 'warn':
            print(
                'WARNING: Overrides are not allowed. Already have method_name=%r. Skipping' %
                (method_name))
            return
        elif allow_override == 'override+warn':
            #import utool as ut
            #ut.embed()
            print('WARNING: Overrides are allowed, but dangerous. method_name=%r.' %
                  (method_name))
            print('old_method = %r, im_func=%s' % (old_method, str(old_im_func)))
            print('new_method = %r, im_func=%s' % (new_method, str(new_im_func)))
            print(get_funcglobals(old_im_func)['__name__'])
            print(get_funcglobals(new_im_func)['__name__'])
        # TODO: does this actually decrement the refcount enough?
        del old_method
    setattr(self, method_name, new_method)


def inject_func_as_property(self, func, method_name=None, class_=None):
    """
    WARNING:
        properties are more safely injected using metaclasses

    References:
        http://stackoverflow.com/questions/13850114/dynamically-adding-methods-with-or-without-metaclass-in-python
    """
    if method_name is None:
        method_name = get_funcname(func)
    #new_method = func.__get__(self, self.__class__)
    new_property = property(func)
    setattr(self.__class__, method_name, new_property)


def inject_func_as_unbound_method(class_, func, method_name=None):
    """ This is actually quite simple """
    if method_name is None:
        method_name = get_funcname(func)
    setattr(class_, method_name, func)


def makeForwardingMetaclass(forwarding_dest_getter, whitelist, base_class=object):
    """
    makes a metaclass that overrides __getattr__ and __setattr__ to forward some
    specific attribute references to a specified instance variable
    """
    class ForwardingMetaclass(base_class.__class__):
        def __init__(metaself, name, bases, dct):
            # print('ForwardingMetaclass.__init__():
            #  {forwarding_dest_getter: %r; whitelist: %r}' % (forwarding_dest_getter, whitelist))
            super(ForwardingMetaclass, metaself).__init__(name, bases, dct)
            old_getattr = metaself.__getattribute__
            old_setattr = metaself.__setattr__
            def new_getattr(self, item):
                if item in whitelist:
                    #dest = old_getattr(self, forwarding_dest_name)
                    dest = forwarding_dest_getter(self)
                    try:
                        val = dest.__class__.__getattribute__(dest, item)
                    except AttributeError:
                        val = getattr(dest, item)
                else:
                    val = old_getattr(self, item)
                return val
            def new_setattr(self, name, val):
                if name in whitelist:
                    #dest = old_getattr(self, forwarding_dest_name)
                    dest = forwarding_dest_getter(self)
                    dest.__class__.__setattr__(dest, name, val)
                else:
                    old_setattr(self, name, val)
            metaself.__getattribute__ = new_getattr
            metaself.__setattr__ = new_setattr
    return ForwardingMetaclass


def test_reloading_metaclass():
    r"""
    CommandLine:
        python -m utool.util_class --test-test_reloading_metaclass

    References:
        http://stackoverflow.com/questions/8122734/pythons-imp-reload-function-is-not-working

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_class import *  # NOQA
        >>> result = test_reloading_metaclass()
        >>> print(result)
    """
    import utool as ut
    testdir = ut.ensure_app_cache_dir('utool', 'metaclass_tests')
    testfoo_fpath = ut.unixjoin(testdir, 'testfoo.py')
    # os.chdir(testdir)
    #with ut.ChdirContext(testdir, stay=ut.inIPython()):
    with ut.ChdirContext(testdir):
        foo_code1 = ut.codeblock(
            r'''
            # STARTBLOCK
            import utool as ut
            import six


            @six.add_metaclass(ut.ReloadingMetaclass)
            class Foo(object):
                def __init__(self):
                    pass

            spamattr = 'version1'
            # ENDBLOCK
            '''
        )
        foo_code2 = ut.codeblock(
            r'''
            # STARTBLOCK
            import utool as ut
            import six


            @six.add_metaclass(ut.ReloadingMetaclass)
            class Foo(object):
                def __init__(self):
                    pass

                def bar(self):
                    return 'spam'

            eggsattr = 'version2'
            # ENDBLOCK
            '''
        )
        # Write a testclass to disk
        ut.delete(testfoo_fpath)
        ut.write_to(testfoo_fpath, foo_code1, verbose=True)
        testfoo = ut.import_module_from_fpath(testfoo_fpath)
        #import testfoo
        foo = testfoo.Foo()
        print('foo = %r' % (foo,))
        assert not hasattr(foo, 'bar'), 'foo should not have a bar attr'
        ut.delete(testfoo_fpath + 'c')  # remove the pyc file because of the identical creation time
        ut.write_to(testfoo_fpath, foo_code2, verbose=True)
        assert not hasattr(foo, 'bar'), 'foo should still not have a bar attr'
        foo.rrr()
        assert foo.bar() == 'spam'
        ut.delete(testfoo_fpath)
        print('Reloading worked nicely')


class ReloadingMetaclass(type):
    """
    Classes with this metaclass will be able to reload themselves
    on a per-instance basis using the rrr function.

    If the functions _on_reload and _initialize_self exist
    they will be called after and before reload respectively. Any
    inject_instance functions should be handled there.

    SeeAlso:
        test_reloading_metaclass - shows a working example of this doctest

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_class import *  # NOQA
        >>> import utool as ut
        >>> @six.add_metaclass(ut.ReloadingMetaclass)
        >>> class Foo(object):
        ...     def __init__(self):
        ...         pass
        >>> # You can edit foo on disk and call rrr in ipython
        >>> # if you add a new function to it
        >>> foo = Foo()
        >>> # This will not work as a doctests because
        >>> # Foo's parent module will be __main__ but
        >>> # there will be no easy way to write to it.
        >>> # This does work when you run from ipython
        >>> @six.add_metaclass(ut.ReloadingMetaclass)
        >>> class Foo(object):
        ...     def __init__(self):
        ...         pass
        ...     def bar(self):
        ...         return 'spam'
        >>> foo.rrr()
        >>> result = foo.bar()
        >>> print(result)
        spam
    """
    def __init__(metaself, name, bases, dct):
        super(ReloadingMetaclass, metaself).__init__(name, bases, dct)
        metaself.rrr = reload_class


def reloading_meta_metaclass_factory(BASE_TYPE=type):
    """ hack for pyqt """
    class ReloadingMetaclass2(BASE_TYPE):
        def __init__(metaself, name, bases, dct):
            super(ReloadingMetaclass2, metaself).__init__(name, bases, dct)
            #print('Making rrr for %r' % (name,))
            metaself.rrr = reload_class
    return ReloadingMetaclass2


def reload_class(self, verbose=True, reload_module=True):
    """
    special class reloading function
    This function is often injected as rrr of classes
    """
    import utool as ut
    verbose = verbose or VERBOSE_CLASS
    classname = self.__class__.__name__
    try:
        modname = self.__class__.__module__
        if verbose:
            print('[class] reloading ' + classname + ' from ' + modname)
        # --HACK--
        if hasattr(self, '_on_reload'):
            if verbose > 1:
                print('[class] calling _on_reload for ' + classname)
            self._on_reload()
        elif verbose > 1:
            print('[class] ' + classname + ' does not have an _on_reload function')

        # Do for all inheriting classes
        def find_base_clases(_class, find_base_clases=None):
            class_list = []
            for _baseclass in _class.__bases__:
                parents = find_base_clases(_baseclass, find_base_clases)
                class_list.extend(parents)
            if _class is not object:
                class_list.append(_class)
            return class_list

        head_class = self.__class__
        # Determine if parents need reloading
        class_list = find_base_clases(head_class, find_base_clases)
        # HACK
        ignore = {HashComparable2}
        class_list = [_class for _class in class_list
                      if _class not in ignore]
        for _class in class_list:
            if verbose:
                print('[class] reloading parent ' + _class.__name__ +
                      ' from ' + _class.__module__)
            if _class.__module__ == '__main__':
                # Attempt to find the module that is the main module
                # This may be very hacky and potentially break
                main_module_ = sys.modules[_class.__module__]
                main_modname = ut.get_modname_from_modpath(main_module_.__file__)
                module_ = sys.modules[main_modname]
            else:
                module_ = sys.modules[_class.__module__]
            if hasattr(module_, 'rrr'):
                if reload_module:
                    module_.rrr(verbose=verbose)
            else:
                if reload_module:
                    import imp
                    if verbose:
                        print('[class] reloading ' + _class.__module__ + ' with imp')
                    try:
                        imp.reload(module_)
                    except (ImportError, AttributeError):
                        print('[class] fallback reloading ' + _class.__module__ +
                              ' with imp')
                        # one last thing to try. probably used ut.import_module_from_fpath
                        # when importing this module
                        imp.load_source(module_.__name__, module_.__file__)
            # Reset class attributes
            _newclass = getattr(module_, _class.__name__)
            reload_class_methods(self, _newclass, verbose=verbose)

        # --HACK--
        # TODO: handle injected definitions
        if hasattr(self, '_initialize_self'):
            if verbose > 1:
                print('[class] calling _initialize_self for ' + classname)
            self._initialize_self()
        elif verbose > 1:
            print('[class] ' + classname + ' does not have an _initialize_self function')
    except Exception as ex:
        ut.printex(ex, 'Error Reloading Class', keys=[
            'modname', 'module', 'class_', 'class_list', 'self', ])
        raise


def reload_class_methods(self, class_, verbose=True):
    """
    rebinds all class methods

    Args:
        self (object): class instance to reload
        class_ (type): type to reload as

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_class import *  # NOQA
        >>> self = '?'
        >>> class_ = '?'
        >>> result = reload_class_methods(self, class_)
        >>> print(result)
    """
    if verbose:
        print('[util_class] Reloading self=%r as class_=%r' % (self, class_))
    self.__class__ = class_
    for key in dir(class_):
        # Get unbound reloaded method
        func = getattr(class_, key)
        if isinstance(func, types.MethodType):
            # inject it into the old instance
            inject_func_as_method(self, func, class_=class_,
                                  allow_override=True,
                                  verbose=verbose)


def compare_instance(op, self, other):
    hash1 = self.__hash__()
    if other.__hash__ is not None:
        hash2 = hash(other)
    else:
        hash2 = other
    try:
        return op(hash1, hash2)
    except Exception as ex:
        import utool as ut
        ut.printex(ex, 'could not compare hash1 to hash2', keys=['hash1', 'hash2'])
        raise


def get_comparison_methods():
    """ makes methods for >, <, =, etc... """
    method_list = []
    def _register(func):
        method_list.append(func)
        return func

    # Comparison operators for sorting and uniqueness
    @_register
    def __lt__(self, other):
        return compare_instance(op.lt, self, other)

    @_register
    def __le__(self, other):
        return compare_instance(op.le, self, other)

    @_register
    def __eq__(self, other):
        return compare_instance(op.eq, self, other)

    @_register
    def __ne__(self, other):
        return compare_instance(op.ne, self, other)

    @_register
    def __gt__(self, other):
        return compare_instance(op.gt, self, other)

    @_register
    def __ge__(self, other):
        return compare_instance(op.ge, self, other)

    return method_list


def get_comparison_operators():
    import operator
    opdict = {
        'not': operator.not_,
        'is': operator.is_,
        'is not': operator.is_not,
        'contains': operator.contains,
        'in': lambda a, b: operator.contains(b, a),
        'not in': lambda a, b: not operator.contains(b, a),
        '!=': operator.ne,
        '==': operator.eq,
        '>=': operator.ge,
        '<=': operator.le,
        '>': operator.gt,
        '<': operator.lt,
    }
    for k, v in list(opdict.items()):
        op = getattr(v, '__name__')
        if op not in opdict and op != '<lambda>':
            opdict[op] = v
    return opdict


class HashComparableMetaclass(type):
    """
    Defines extra methods for Configs

    FIXME: this breaks in python3 because
    anything that overwrites hash overwrites inherited __eq__

    https://docs.python.org/3.6/reference/datamodel.html#object.__hash__
    """
    def __new__(cls, name, bases, dct):
        """
        Args:
            cls (type): meta
            name (str): classname
            supers (list): bases
            dct (dict): class dictionary
        """
        method_list = get_comparison_methods()
        for func in method_list:
            if get_funcname(func) not in dct:
                funcname = get_funcname(func)
                dct[funcname] = func
            else:
                funcname = get_funcname(func)
                dct['meta_' + funcname] = func
            #ut.inject_func_as_method(metaself, func)
        return type.__new__(cls, name, bases, dct)


@six.add_metaclass(HashComparableMetaclass)
class HashComparable(object):
    pass


class HashComparable2(object):
    def __lt__(self, other):
        return compare_instance(op.lt, self, other)

    def __le__(self, other):
        return compare_instance(op.le, self, other)

    def __eq__(self, other):
        return compare_instance(op.eq, self, other)

    def __ne__(self, other):
        return compare_instance(op.ne, self, other)

    def __gt__(self, other):
        return compare_instance(op.gt, self, other)

    def __ge__(self, other):
        return compare_instance(op.ge, self, other)


def reloadable_class(cls):
    """
    convinience decorator instead of @six.add_metaclass(ReloadingMetaclass)
    """
    return six.add_metaclass(ReloadingMetaclass)(cls)


class KwargsWrapper(Mapping):
    """
    Allows an arbitrary object attributes to be passed as a **kwargs
    argument
    """
    def __init__(self, obj):
        self.obj = obj

    def __getitem__(self, key):
        return self.obj.__dict__[key]

    def __iter__(self):
        return iter(self.obj.__dict__)

    def __len__(self):
        return len(self.obj.__dict__)


def remove_private_obfuscation(self):
    """
    removes the python obfuscation of class privates so they can be executed as
    they appear in class source. Useful when playing with IPython.
    """
    classname = self.__class__.__name__
    attrlist = [attr for attr in dir(self) if attr.startswith('_' + classname + '__')]
    for attr in attrlist:
        method = getattr(self, attr)
        truename = attr.replace('_' + classname + '__', '__')
        setattr(self, truename, method)


def get_classname(class_, local=False):
    r"""
    Args:
        class_ (type):
        local (bool): (default = False)

    Returns:
        str: classname

    CommandLine:
        python -m utool.util_class --exec-get_classname --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_class import *  # NOQA
        >>> import utool as ut
        >>> class_ = ReloadingMetaclass
        >>> local = False
        >>> assert get_classname(class_, local) == 'utool.util_class.ReloadingMetaclass'
        >>> assert get_classname(class_, local=True) == 'ReloadingMetaclass'
    """
    if not local:
        classname = class_.__module__ + '.' + class_.__name__
    else:
        classname = class_.__name__
    return classname


if __name__ == '__main__':
    """
    python -c "import utool; utool.doctest_funcs(utool.util_class, allexamples=True)"
    python -m utool.util_class --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()
    import utool as ut
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from utool import util_inject
import six
import collections
print, rrr, profile = util_inject.inject2(__name__)


def get_tablenames(cur):
    """ Conveinience: """
    cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
    tablename_list_ = cur.fetchall()
    tablename_list = [str(tablename[0]) for tablename in tablename_list_ ]
    return tablename_list

SQLColumnRichInfo = collections.namedtuple('SQLColumnRichInfo', ('column_id', 'name', 'type_', 'notnull', 'dflt_value', 'pk'))


def get_table_columns(cur, tablename, exclude_columns=[]):
    import utool as ut
    colnames_ = ut.get_table_columnname_list(cur, tablename)
    colnames = tuple([colname for colname in colnames_ if colname not in exclude_columns])
    row_list = ut.get_table_rows(cur, tablename, colnames, unpack=False)
    column_list = zip(*row_list)
    return column_list


def get_table_csv(cur, tablename, exclude_columns=[]):
    """ Conveinience: Converts a tablename to csv format

    Args:
        tablename (str):
        exclude_columns (list):

    Returns:
        str: csv_table

    CommandLine:
        python -m ibeis.control.SQLDatabaseControl --test-get_table_csv

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.control.SQLDatabaseControl import *  # NOQA
        >>> # build test data
        >>> import ibeis
        >>> ibs = ibeis.opendb('testdb1')
        >>> db = ibs.db
        >>> tablename = ibeis.const.NAME_TABLE
        >>> exclude_columns = []
        >>> # execute function
        >>> csv_table = db.get_table_csv(tablename, exclude_columns)
        >>> # verify results
        >>> result = str(csv_table)
        >>> print(result)
    """
    import utool as ut
    colnames_ = ut.get_table_columnname_list(cur, tablename)
    colnames = tuple([colname for colname in colnames_ if colname not in exclude_columns])
    row_list = ut.get_table_rows(cur, tablename, colnames, unpack=False)
    column_list = zip(*row_list)
    #=None, column_list=[], header='', column_type=None
    #import utool as ut
    #column_list, column_names = db.get_table_column_data(tablename, exclude_columns)
    # remove column prefix for more compact csvs
    column_lbls = [name.replace(tablename[:-1] + '_', '') for name in colnames]
    #header = db.get_table_csv_header(tablename)
    header = ''
    csv_table = ut.make_csv_table(column_list, column_lbls, header)
    return csv_table


def get_table_columnname_list(cur, tablename):
    colinfo_list_ = get_table_columninfo_list(cur, tablename)
    return [info[1] for info in colinfo_list_]


def get_table_columninfo_list(cur, tablename):
    """
    Args:
        tablename (str): table name

    Returns:
        column_list : list of tuples with format:
            (
                [0] column_id  : id of the column
                [1] name       : the name of the column
                [2] type_      : the type of the column (TEXT, INT, etc...)
                [3] notnull    : 0 or 1 if the column can contains null values
                [4] dflt_value : the default value
                [5] pk         : 0 or 1 if the column partecipate to the primary key
            )

    References:
        http://stackoverflow.com/questions/17717829/how-to-get-column-names-from-a-table-in-sqlite-via-pragma-net-c

    CommandLine:
        python -m utool.util_sqlite --test-get_table_columninfo_list

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_sqlite import *  # NOQA
    """
    cur.execute('PRAGMA TABLE_INFO("{tablename}")'.format(tablename=tablename))
    colinfo_list = cur.fetchall()
    colrichinfo_list = [SQLColumnRichInfo(*colinfo) for colinfo in colinfo_list]
    return colrichinfo_list


def get_primary_columninfo(cur, tablename):
    colinfo_list_ = get_table_columninfo_list(cur, tablename)
    colinfo_list = [colinfo for colinfo in colinfo_list_ if colinfo.pk]
    return colinfo_list


def get_nonprimary_columninfo(cur, tablename):
    colinfo_list_ = get_table_columninfo_list(cur, tablename)
    colinfo_list = [colinfo for colinfo in colinfo_list_ if not colinfo.pk]
    return colinfo_list


def get_table_num_rows(cur, tablename):
    cur.execute('SELECT COUNT(*) FROM {tablename}'.format(tablename=tablename))
    num_rows = cur.fetchall()[0][0]
    return num_rows


def get_table_column(cur, tablename, colname):
    """ Conveinience: """
    return get_table_rows(cur, tablename, colname)


def get_table_rows(cur, tablename, colnames, where=None, params=None, unpack=True):
    import utool as ut
    want_single_column = isinstance(colnames, six.string_types)
    want_single_param = params is not None and not ut.isiterable(params)
    #isinstance(params, six.string_types)
    if want_single_column:
        colnames = (colnames,)
    if colnames is not None and colnames != '*':
        assert isinstance(colnames, tuple), 'colnames must be a tuple'
        colnames_str = ', '.join(colnames)
    else:
        colnames_str = '*'
    #if isinstance(colnames, six.string_types):
    #    colnames = (colnames,)
    fmtdict = {
        'tablename'     : tablename,
        'colnames'    : colnames_str,
        'orderby'     : '',
    }

    #ORDER BY rowid ASC
    if where is None:
        operation_fmt = '''
        SELECT {colnames}
        FROM {tablename}
        {orderby}
        '''
    else:
        fmtdict['where_clause'] = where
        operation_fmt = '''
        SELECT {colnames}
        FROM {tablename}
        WHERE {where_clause}
        {orderby}
        '''
    operation_str = operation_fmt.format(**fmtdict)
    if params is None:
        cur.execute(operation_str)
        val_list = cur.fetchall()
    elif want_single_param:
        cur.execute(operation_str, (params,))
        val_list = cur.fetchall()
    else:
        # Execute many
        def executemany_scalar_generator(operation_str, params):
            for param in params:
                cur.execute(operation_str, param)
                vals = cur.fetchall()
                #assert len(vals) == 1, 'vals=%r, len(vals)=%r' % (vals, len(vals))
                yield vals
        val_list = list(executemany_scalar_generator(operation_str, params))

    if unpack:
        if want_single_column:
            # want a single value per parameter
            val_list = [val[0] for val in val_list]

        if want_single_param:
            # wants a single parameter
            assert len(val_list) == 1
            val_list = val_list[0]
    return val_list


def print_database_structure(cur):
    import utool as ut
    tablename_list = ut.get_tablenames(cur)
    colinfos_list = [ut.get_table_columninfo_list(cur, tablename) for tablename in tablename_list]
    numrows_list = [ut.get_table_num_rows(cur, tablename) for tablename in tablename_list]
    for tablename, colinfo_list, num_rows in ut.sortedby(list(zip(tablename_list, colinfos_list, numrows_list)), numrows_list):
        print('+-------------')
        print('tablename = %r' % (tablename,))
        print('num_rows = %r' % (num_rows,))
        #print(ut.repr4(colinfo_list))
        print(ut.repr4(ut.get_primary_columninfo(cur, tablename)))
        print(ut.repr4(ut.get_nonprimary_columninfo(cur, tablename)))
        print('+-------------')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import os
import utool
from os.path import basename


def make_setup(repodir):
    setup_text_fmt_ = '''
    # autogenerated setup.py on {timestamp} for {repodir}
    from __future__ import absolute_import, division, print_function
    import setuptools
    #from utool import util_setup

    INSTALL_REQUIRES = [
    ]

    if __name__ == '__main__':
        setuptools.setup(
            name='{pkgname}',
            packages={packages},
            #packages=util_setup.find_packages(),
            #version='0.0.0.autogen',
            #description='short description',
            #url='',
            #ext_modules=util_setup.find_ext_modules(),
            #cmdclass=util_setup.get_cmdclass(),
            #author=''
            #author_email='',
            #keywords='',
            #install_requires=INSTALL_REQUIRES,
            package_data={{}},
            scripts=[
            ],
            classifiers=[],
        )
    '''
    setup_text_fmt = utool.unindent(setup_text_fmt_)
    timestamp = utool.get_timestamp()
    pkgname = basename(repodir)
    packages = utool.ls_moduledirs(repodir, full=False)
    print(pkgname)
    setup_text = setup_text_fmt.format(
        packages=packages,
        repodir=repodir,
        timestamp=timestamp,
        pkgname=pkgname,
    )
    return setup_text


if __name__ == '__main__':
    writeflag = utool.get_argflag(('--write', '-w'))
    overwriteflag = utool.get_argflag(('--yes', '-y'))
    repodir = utool.unixpath(os.getcwd())
    print('[utool] making setup.py for: %r' % repodir)
    setup_text = make_setup(repodir)
    if writeflag:
        setup_fpath = utool.unixjoin(repodir, 'setup.py')
        if utool.checkpath(setup_fpath):
            confirm_flag = overwriteflag
        else:
            confirm_flag = True
        if confirm_flag:
            utool.write_to(setup_fpath, setup_text)
        else:
            print('setup.py file exists not writing')
    else:
        print(setup_text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Cleaning script for the output of utool profiling

Removes profiled output of code that never ran
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from six.moves import range, cStringIO
import six
import operator
import utool as ut

"""
input_fname = 'raw_profile.dev.py.2014-09-23_18-28-57.raw.prof'
input_fname = 'raw_profile.qt_inc_automatch.py.2014-12-23_11-31-53.raw.prof'
"""


class Profiler(object):
    def __init__(self):
        import line_profiler
        self._lineprof = line_profiler.LineProfiler()

    def __call__(self, func):
        wrapper = self._lineprof(func)
        # wrapper = ut.preserve_sig(wrapper, func)
        return wrapper

    def get_output(self):
        file_ = cStringIO()
        self._lineprof.print_stats(stream=file_, stripzeros=True)
        file_.seek(0)
        text =  file_.read()
        output_text, summary_text = ut.clean_line_profile_text(text)
        return output_text


def make_profiler():
    import line_profiler
    profile = line_profiler.LineProfiler()
    return profile


def get_profile_text(profile):
    file_ = cStringIO()
    profile.print_stats(stream=file_, stripzeros=True)
    file_.seek(0)
    text =  file_.read()
    output_text, summary_text = ut.clean_line_profile_text(text)
    return output_text, summary_text


def dump_profile_text():
    import utool as ut
    print("Dumping Profile Information")
    profile = ut.PROFILE_FUNC
    try:
        output_text, summary_text = get_profile_text(profile)
    except AttributeError:
        print('profile is not on')
    else:
        #profile.dump_stats('out.lprof')
        print(summary_text)
        ut.writeto('profile_output.txt', output_text + '\n' + summary_text)
        ut.writeto('profile_output.%s.txt' % (ut.get_timestamp()),
                   output_text + '\n' + summary_text)


def __dbg_list(profile_block_list):
    for item in profile_block_list:
        newline = item.find('\n')
        twoline = item[newline + 1:].find('\n')
        head = item[0:(newline + twoline)]
        print(head)


def parse_rawprofile_blocks(text):
    """
    Split the file into blocks along delimters and and put delimeters back in
    the list
    """
    # The total time reported in the raw output is from pystone not kernprof
    # The pystone total time is actually the average time spent in the function
    delim = 'Total time: '
    delim2 = 'Pystone time: '
    #delim = 'File: '
    profile_block_list = ut.regex_split('^' + delim, text)
    for ix in range(1, len(profile_block_list)):
        profile_block_list[ix] = delim2 + profile_block_list[ix]
    return profile_block_list


def get_block_totaltime(block):
    time_line = ut.regex_search('Pystone time: [0-9.]* s', block)
    time_str  = ut.regex_search('[0-9.]+', time_line)
    if time_str is not None:
        return float(time_str)
    else:
        return None


def get_block_id(block):
    import re
    fpath_regex = ut.named_field('fpath', r'\S+')
    funcname_regex = ut.named_field('funcname', r'\S+')
    lineno_regex = ut.named_field('lineno', '[0-9]+')

    fileline_regex = 'File: ' + fpath_regex + '$'
    funcline_regex = 'Function: ' + funcname_regex + ' at line ' + lineno_regex + '$'
    fileline_match = re.search(fileline_regex, block, flags=re.MULTILINE)
    funcline_match = re.search(funcline_regex, block, flags=re.MULTILINE)
    if fileline_match is not None and funcline_match is not None:
        fpath    = fileline_match.groupdict()['fpath']
        funcname = funcline_match.groupdict()['funcname']
        lineno   = funcline_match.groupdict()['lineno']
        block_id = funcname + ':' + fpath + ':' + lineno
    else:
        block_id = 'None:None:None'
    return block_id


def parse_timemap_from_blocks(profile_block_list):
    """
    Build a map from times to line_profile blocks
    """
    prefix_list = []
    timemap = ut.ddict(list)
    for ix in range(len(profile_block_list)):
        block = profile_block_list[ix]
        total_time = get_block_totaltime(block)
        # Blocks without time go at the front of sorted output
        if total_time is None:
            prefix_list.append(block)
        # Blocks that are not run are not appended to output
        elif total_time != 0:
            timemap[total_time].append(block)
    return prefix_list, timemap


def get_summary(profile_block_list, maxlines=20):
    """
    References:
        https://github.com/rkern/line_profiler
    """
    time_list = [get_block_totaltime(block) for block in profile_block_list]
    time_list = [time if time is not None else -1 for time in time_list]
    blockid_list = [get_block_id(block) for block in profile_block_list]
    sortx = ut.list_argsort(time_list)
    sorted_time_list = ut.take(time_list, sortx)
    sorted_blockid_list = ut.take(blockid_list, sortx)

    aligned_blockid_list = ut.util_str.align_lines(sorted_blockid_list, ':')
    summary_lines = [('%6.2f seconds - ' % time) + line
                     for time, line in
                     zip(sorted_time_list, aligned_blockid_list)]
    #summary_header = ut.codeblock(
    #    '''
    #    CLEANED PROFILE OUPUT

    #    The Pystone timings are not from kernprof, so they may include kernprof
    #    overhead, whereas kernprof timings do not (unless the line being
    #    profiled is also decorated with kernrof)

    #    The kernprof times are reported in Timer Units

    #    ''')
    summary_lines_ = ut.listclip(summary_lines, maxlines, fromback=True)
    summary_text = '\n'.join(summary_lines_)
    return summary_text


def fix_rawprofile_blocks(profile_block_list):
    # TODO: finish function. should multiply times by
    # Timer unit to get true second profiling
    #profile_block_list_new = []
    for block in profile_block_list:
        block_lines = block.split('\n')
        sep = ['=' * 62]
        def split_block_at_sep(block_lines, sep):
            for pos, line in enumerate(block_lines):
                if line.find(sep) == 0:
                    pos += 1
                    header_lines = block_lines[:pos]
                    body_lines = block_lines[pos:]
                    return header_lines, body_lines
            return block_lines, None
        header_lines, body_lines = split_block_at_sep(block_lines, sep)


def clean_line_profile_text(text):
    """
    Sorts the output from line profile by execution time
    Removes entries which were not run
    """
    #
    profile_block_list = parse_rawprofile_blocks(text)
    #profile_block_list = fix_rawprofile_blocks(profile_block_list)
    #---
    # FIXME can be written much nicer
    prefix_list, timemap = parse_timemap_from_blocks(profile_block_list)
    # Sort the blocks by time
    sorted_lists = sorted(six.iteritems(timemap), key=operator.itemgetter(0))
    newlist = prefix_list[:]
    for key, val in sorted_lists:
        newlist.extend(val)
    # Rejoin output text
    output_text = '\n'.join(newlist)
    #---
    # Hack in a profile summary
    summary_text = get_summary(profile_block_list)
    output_text = output_text
    return output_text, summary_text


def clean_lprof_file(input_fname, output_fname=None):
    """ Reads a .lprof file and cleans it """
    # Read the raw .lprof text dump
    text = ut.read_from(input_fname)
    # Sort and clean the text
    output_text = clean_line_profile_text(text)
    return output_text


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m utool.util_profile
        python -m utool.util_profile --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import sys
import os
from os.path import split, exists, join, dirname
from utool import util_inject
from utool._internal import meta_util_arg
print, rrr, profile = util_inject.inject2(__name__)


def in_virtual_env():
    """
    returns True if you are running inside a python virtual environment.
    (DOES NOT WORK IF IN IPYTHON AND USING A VIRTUALENV)

    sys.prefix gives the location of the virtualenv

    Notes:
        It seems IPython does not respect virtual environments properly.
        TODO: find a solution
        http://stackoverflow.com/questions/7335992/ipython-and-virtualenv-ignoring-site-packages

    References:
        http://stackoverflow.com/questions/1871549/python-determine-if-running-inside-virtualenv

    CommandLine:
        python -m utool.util_sysreq in_virtual_env

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_sysreq import *  # NOQA
        >>> import utool as ut
        >>> result = in_virtual_env()
        >>> print(result)
    """
    import sys
    has_venv = False
    if hasattr(sys, 'real_prefix'):
        # For virtualenv module
        has_venv = True
    elif hasattr(sys, 'base_prefix'):
        # For venv module
        has_venv = sys.base_prefix != sys.prefix
    return has_venv


def get_site_packages_dir():
    """
    CommandLine:
        python -m utool.util_sysreq get_site_packages_dir

    Notes:
        It seems IPython does not respect virtual environments properly.
        TODO: find a solution
        http://stackoverflow.com/questions/7335992/ipython-and-virtualenv-ignoring-site-packages
    """
    import distutils.sysconfig
    return distutils.sysconfig.get_python_lib()


def get_global_dist_packages_dir():
    """
    Attempts to work around virtualenvs and find the system dist_pacakges.
    Essentially this is implmenented as a lookuptable
    """
    import utool as ut
    if not ut.in_virtual_env():
        # Non venv case
        return get_site_packages_dir()
    else:
        candidates = []
        if ut.LINUX:
            import sys
            candidates += [
                '/usr/lib/python%s/dist-packages' % (sys.version[0:3],),
                '/usr/lib/python%s/dist-packages' % (sys.version[0:1],),
            ]
        else:
            raise NotImplementedError()
        for path in candidates:
            if ut.checkpath(path):
                return path


def get_local_dist_packages_dir():
    """
    Attempts to work around virtualenvs and find the system dist_pacakges.
    Essentially this is implmenented as a lookuptable
    """
    import utool as ut
    if not ut.in_virtual_env():
        # Non venv case
        return get_site_packages_dir()
    else:
        candidates = []
        if ut.LINUX:
            candidates += [
                '/usr/local/lib/python2.7/dist-packages',
            ]
        else:
            raise NotImplementedError()
        for path in candidates:
            if ut.checkpath(path):
                return path


def is_running_as_root():
    """
    References:
        http://stackoverflow.com/questions/5721529/running-python-script-as-root-with-sudo-what-is-the-username-of-the-effectiv
        http://stackoverflow.com/questions/2806897/what-is-the-best-practices-for-checking-if-the-user-of-a-python-script-has-root
    """
    return os.getenv('USER') == 'root'


def locate_path(dname, recurse_down=True):
    """ Search for a path """
    tried_fpaths = []
    root_dir = os.getcwd()
    while root_dir is not None:
        dpath = join(root_dir, dname)
        if exists(dpath):
            return dpath
        else:
            tried_fpaths.append(dpath)
        _new_root = dirname(root_dir)
        if _new_root == root_dir:
            root_dir = None
            break
        else:
            root_dir = _new_root
        if not recurse_down:
            break
    msg = 'Cannot locate dname=%r' % (dname,)
    msg = ('\n[sysreq!] Checked: '.join(tried_fpaths))
    print(msg)
    raise ImportError(msg)


def ensure_in_pythonpath(dname):
    dname_list = [split(dpath)[1] for dpath in sys.path]
    if dname not in dname_list:
        dpath = locate_path(dname)
        if meta_util_arg.VERBOSE:
            print('[sysreq] appending %r to PYTHONPATH' % dpath)
        sys.path.append(dpath)
    elif meta_util_arg.DEBUG:
        print('[sysreq] PYTHONPATH has %r' % dname)


def total_purge_developed_repo(repodir):
    r"""
    Outputs commands to help purge a repo

    Args:
        repodir (str): path to developed repository

    CommandLine:
        python -m utool.util_sysreq total_purge_installed_repo --show

    Ignore:
        repodir = ut.truepath('~/code/Lasagne')

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_sysreq import *  # NOQA
        >>> import utool as ut
        >>> repodir = ut.get_argval('--repodir', default=None)
        >>> result = total_purge_installed_repo(repodir)
    """
    assert repodir is not None
    import utool as ut
    import os
    repo = ut.util_git.Repo(dpath=repodir)

    user = os.environ['USER']

    fmtdict = dict(
        user=user,
        modname=repo.modname,
        reponame=repo.reponame,
        dpath=repo.dpath,
        global_site_pkgs=ut.get_global_dist_packages_dir(),
        local_site_pkgs=ut.get_local_dist_packages_dir(),
        venv_site_pkgs=ut.get_site_packages_dir(),
    )

    commands = [_.format(**fmtdict) for _ in [
        'pip uninstall {modname}',
        'sudo -H pip uninstall {modname}',
        'sudo pip uninstall {modname}',
        'easy_install -m {modname}',
        'cd {dpath} && python setup.py develop --uninstall',
        # If they still exist try chowning to current user
        'sudo chown -R {user}:{user} {dpath}',
    ]]
    print('Normal uninstall commands')
    print('\n'.join(commands))

    possible_link_paths = [_.format(**fmtdict) for _ in [
        '{dpath}/{modname}.egg-info',
        '{dpath}/build',
        '{venv_site_pkgs}/{reponame}.egg-info',
        '{local_site_pkgs}/{reponame}.egg-info',
        '{venv_site_pkgs}/{reponame}.egg-info',
    ]]
    from os.path import exists, basename
    existing_link_paths = [path for path in possible_link_paths]
    print('# Delete paths and eggs')
    for path in existing_link_paths:
        if exists(path):
            if ut.get_file_info(path)['owner'] != user:
                print('sudo /bin/rm -rf {path}'.format(path=path))
            else:
                print('/bin/rm -rf {path}'.format(path=path))
        #ut.delete(path)

    print('# Make sure nothing is in the easy install paths')
    easyinstall_paths = [_.format(**fmtdict) for _ in [
        '{venv_site_pkgs}/easy-install.pth',
        '{local_site_pkgs}/easy-install.pth',
        '{venv_site_pkgs}/easy-install.pth',
    ]]
    for path in easyinstall_paths:
        if exists(path):
            easy_install_list = ut.readfrom(path, verbose=False).strip().split('\n')
            easy_install_list_ = [basename(p) for p in easy_install_list]
            index1 = ut.listfind(easy_install_list_, repo.reponame)
            index2 = ut.listfind(easy_install_list_, repo.modname)
            if index1 is not None or index2 is not None:
                print('Found at index1=%r, index=%r' % (index1, index2))
                if ut.get_file_info(path)['owner'] != user:
                    print('sudo gvim {path}'.format(path=path))
                else:
                    print('gvim {path}'.format(path=path))

    checkcmds = [_.format(**fmtdict) for _ in [
        'python -c "import {modname}; print({modname}.__file__)"'
    ]]
    import sys
    assert repo.modname not in sys.modules
    print("# CHECK STATUS")
    for cmd in checkcmds:
        print(cmd)
        #ut.cmd(cmd, verbose=False)


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m utool.util_sysreq
        python -m utool.util_sysreq --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import sys
import six
from os.path import normpath, expanduser, join

OS_TYPE = sys.platform
if six.PY2 and OS_TYPE == 'linux2':
    OS_TYPE = 'linux'  # python2 fix

WIN32  = sys.platform.startswith('win32')
LINUX  = sys.platform.startswith('linux')
DARWIN = sys.platform.startswith('darwin')


def get_resource_dir():
    """
    Returns a directory which should be writable for any application
    """
    #resource_prefix = '~'
    if WIN32:
        dpath_ = '~/AppData/Roaming'
    elif LINUX:
        dpath_ = '~/.config'
    elif DARWIN:
        dpath_  = '~/Library/Application Support'
    else:
        raise AssertionError('unknown os')
    dpath = normpath(expanduser(dpath_))
    return dpath


def get_app_resource_dir(*args, **kwargs):
    """ Returns a writable directory for an application
    Input: appname - the name of the application
           *args, - any other subdirectories may be specified
    """
    if len(args) == 0:
        raise AssertionError('Missing appname. The first argument the '
                             'application name')
    return join(get_resource_dir(), *args)



def check_jedi_can_read_googlestyle():
    import jedi
    import utool as ut
    source1 = ut.codeblock(
        r'''
        # STARTBLOCK
        def spam(data):
            r"""
            Args:
                data (utool.ColumnLists): a column list objct
            """
            data.
        # ENDBLOCK
        '''
    )
    source2 = ut.codeblock(
        r'''
        # STARTBLOCK
        def spam(ibs, bar):
            r"""
            Args:
                ibs (ibeis.IBEISController): an object
            """
            import jedi
            jedi.n
            x = ''
            x.l
            ibs.d
            bar.d
        # ENDBLOCK
        '''
    )

    print('\n---testing jedi with utool.ColumnLists')
    self = script = jedi.Script(source1, line=7, column=8)  # NOQA
    completions = script.completions()  # NOQA
    print('completions = %r' % (completions,))
    vartype = script.goto_definitions()
    print('vartype = %r' % (vartype,))

    print('\n---testing jedi with ibeis.IBEISController')
    script = jedi.Script(source2, line=10)
    script.completions()
    # Find the variable type of argument
    self = script = jedi.Script(source2, line=11, column=7)  # NOQA
    completions = script.completions()  # NOQA
    print('completions = %r' % (completions,))
    vartype = script.goto_definitions()
    print('vartype = %r' % (vartype,))

    print('\n---testing jedi with undefined object bar')
    self = script = jedi.Script(source2, line=12, column=7)  # NOQA
    vartype = script.goto_definitions()  # NOQA
    print('vartype = %r' % (vartype,))
    vardefs = script.goto_assignments()  # NOQA
    print('vardefs = %r' % (vardefs,))
    # foodef, = jedi.names(source2)
    # foomems = foodef.defined_names()
    # xdef = foomems[2]


def _insource_jedi_vim_test(data, ibs):
    """
    If jedi-vim supports google style docstrings you should be able to
    autocomplete ColumnLists methods for `data`

    Args:
        data (utool.ColumnLists): a column list objct
        ibs (ibeis.IBEISController): an object
    """
    # TESTME: type a dot and tab. Hopefully autocomplete will happen.
    data
    ibs
    import utool as ut
    xdata = ut.ColumnLists()
    xdata
    import ibeis
    xibs = ibeis.IBEISController()
    xibs


def check_jedi_closures():
    import jedi
    import textwrap
    source = textwrap.dedent(
        r'''
        def foo(data):
            import matplotlib
            import matplotlib

            def get_default_edge_data(graph, edge):
                data = graph.get_edge_data(*edge)
                if data is None:
                    if len(edge) == 3 and edge[2] is not None:
                        data = graph.get_edge_data(edge[0], edge[1], int(edge[2]))
                    else:
                        data = graph.get_edge_data(edge[0], edge[1])
                if data is None:
                    data = {}
                return data

            import matplotlib
        '''
    )
    print('SOURCE: ')
    print(source)
    lines = source.split('\n')
    test_positions = []
    for row, line in enumerate(lines):
        print('line = %r' % (line,))
        pat = 'import '
        import_pos = line.find(pat)
        if import_pos >= 0:
            module_pos = import_pos + len(pat) + 1
            print('Adding to tests ' + line[module_pos:])
            column = module_pos
            test_positions.append((row, column))

    for row, column in test_positions:
        script = jedi.Script(source, line=3, column=12)
        definitions = script.goto_definitions()
        print('----------')
        print('Script@(%r, %r):' % (row, column))
        print(lines[row])
        print(' ' * column + '^')
        print('definitions = %r' % (definitions,))
        print('----------')

    # script = jedi.Script(source, line=4, column=12)
    # definitions = script.goto_definitions()
    # print('definitions = %r' % (definitions,))

    # script = jedi.Script(source, line=num_lines - 2, column=12)
    # definitions = script.goto_definitions()
    # print('definitions = %r' % (definitions,))


def check_jedi_utool():
    import jedi
    import textwrap
    source = textwrap.dedent(
        r'''
        import utool as ut
        ut.'''
    )
    script = jedi.Script(source)

    definitions = script.goto_definitions()
    print('definitions = %r' % (definitions,))

    completions = script.completions()  # NOQA
    print('completions = %r' % (completions,))

    vardefs = script.goto_assignments()  # NOQA
    print('vardefs = %r' % (vardefs,))


if __name__ == '__main__':
    r"""
    CommandLine:
        export PYTHONPATH=$PYTHONPATH:/home/joncrall/code/utool/utool/util_scripts
        python ~/code/utool/utool/util_scripts/check_jedi.py
    """
    # check_jedi_can_read_googlestyle()
    # check_jedi_closures()
    check_jedi_utool()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import six
import re
import operator
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


def modify_tags(tags_list, direct_map=None, regex_map=None, regex_aug=None,
                delete_unmapped=False, return_unmapped=False,
                return_map=False):
    import utool as ut
    tag_vocab = ut.unique(ut.flatten(tags_list))
    alias_map = ut.odict()
    if regex_map is not None:
        alias_map.update(**ut.build_alias_map(regex_map, tag_vocab))
    if direct_map is not None:
        alias_map.update(ut.odict(direct_map))

    new_tags_list = tags_list
    new_tags_list = ut.alias_tags(new_tags_list, alias_map)

    if regex_aug is not None:
        alias_aug = ut.build_alias_map(regex_aug, tag_vocab)
        aug_tags_list = ut.alias_tags(new_tags_list, alias_aug)
        new_tags_list = [ut.unique(t1 + t2) for t1, t2 in zip(new_tags_list, aug_tags_list)]

    unmapped = list(set(tag_vocab) - set(alias_map.keys()))
    if delete_unmapped:
        new_tags_list = [ut.setdiff(tags, unmapped) for tags in new_tags_list]

    toreturn = None
    if return_map:
        toreturn = (alias_map,)

    if return_unmapped:
        toreturn = toreturn + (unmapped,)

    if toreturn is None:
        toreturn = new_tags_list
    else:
        toreturn = (new_tags_list,) + toreturn
    return toreturn


def tag_coocurrence(tags_list):
    import utool as ut
    co_occur_list = []
    for tags in tags_list:
        for combo in ut.combinations(tags, 2):
            key = tuple(sorted(combo))
            co_occur_list.append(key)
    co_occur = ut.dict_hist(co_occur_list, ordered=True)
    #        co_occur[key] += 1
    #co_occur = ut.odict(co_occur)
    return co_occur


def tag_hist(tags_list):
    import utool as ut
    return ut.dict_hist(ut.flatten(tags_list), ordered=True)


def build_alias_map(regex_map, tag_vocab):
    """
    Constructs explicit mapping. Order of items in regex map matters.
    Items at top are given preference.

    Example:
        >>> # DISABLE_DOCTEST
        >>> tags_list = [['t1', 't2'], [], ['t3'], ['t4', 't5']]
        >>> tag_vocab = ut.flat_unique(*tags_list)
        >>> regex_map = [('t[3-4]', 'A9'), ('t0', 'a0')]
        >>> unmapped = list(set(tag_vocab) - set(alias_map.keys()))
    """
    import utool as ut
    import re
    alias_map = ut.odict([])
    for pats, new_tag in reversed(regex_map):
        pats = ut.ensure_iterable(pats)
        for pat in pats:
            flags = [re.match(pat, t) for t in tag_vocab]
            for old_tag in ut.compress(tag_vocab, flags):
                alias_map[old_tag] = new_tag
    identity_map = ut.take_column(regex_map, 1)
    for tag in ut.filter_Nones(identity_map):
        alias_map[tag] = tag
    return alias_map


def alias_tags(tags_list, alias_map):
    """
    update tags to new values

    Args:
        tags_list (list):
        alias_map (list): list of 2-tuples with regex, value

    Returns:
        list: updated tags

    CommandLine:
        python -m utool.util_tags alias_tags --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tags import *  # NOQA
        >>> import utool as ut
        >>> tags_list = [['t1', 't2'], [], ['t3'], ['t4', 't5']]
        >>> ut.build_alias_map()
        >>> result = alias_tags(tags_list, alias_map)
        >>> print(result)
    """
    def _alias_dict(tags):
        tags_ = [alias_map.get(t, t) for t in tags]
        return list(set([t for t in tags_ if t is not None]))
    tags_list_ = [_alias_dict(tags) for tags in tags_list]
    return tags_list_
    # def _fix_tags(tags):
    #     return {six.text_type(t.lower()) for t in tags}
    # tags_list_ = list(map(_fix_tags, tags_list))
    # re_list = [re.compile(pat) for pat, val in alias_map]
    # val_list = ut.take_column(alias_map, 0)
    # def _alias_regex(tags):
    #     new_tags = 0
    #     for t in tags:
    #         matched = [re_.match(t) is not None for re_ in re_list]
    #         matched_idx = ut.where(matched)
    #         assert len(matched_idx) <= 1, 'more than one tag in %r matched pattern' % (tags,)
    #         if len(matched_idx) > 0:
    #             repl_tags = ut.take(val_list, matched_idx)
    #             new_tags.extend(repl_tags)
    #         else:
    #             new_tags.append(t)
    #     return new_tags
    # # tags_list_ = [_alias_regex(tags) for tags in tags_list_]
    # return tags_list_


def filterflags_general_tags(tags_list, has_any=None, has_all=None,
                             has_none=None, min_num=None, max_num=None,
                             any_startswith=None, any_endswith=None,
                             in_any=None, any_match=None, none_match=None,
                             logic='and', ignore_case=True):
    r"""
    maybe integrate into utool? Seems pretty general

    Args:
        tags_list (list):
        has_any (None): (default = None)
        has_all (None): (default = None)
        min_num (None): (default = None)
        max_num (None): (default = None)

    Notes:
        in_any should probably be ni_any

    TODO: make this function more natural

    CommandLine:
        python -m utool.util_tags --exec-filterflags_general_tags
        python -m utool.util_tags --exec-filterflags_general_tags:0  --helpx
        python -m utool.util_tags --exec-filterflags_general_tags:0
        python -m utool.util_tags --exec-filterflags_general_tags:0  --none_match n
        python -m utool.util_tags --exec-filterflags_general_tags:0  --has_none=n,o
        python -m utool.util_tags --exec-filterflags_general_tags:1
        python -m utool.util_tags --exec-filterflags_general_tags:2

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_tags import *  # NOQA
        >>> import utool as ut
        >>> tags_list = [['v'], [], ['P'], ['P', 'o'], ['n', 'o'], [], ['n', 'N'], ['e', 'i', 'p', 'b', 'n'], ['q', 'v'], ['n'], ['n'], ['N']]
        >>> kwargs = ut.argparse_dict(ut.get_kwdefaults2(filterflags_general_tags), type_hint=list)
        >>> print('kwargs = %r' % (kwargs,))
        >>> flags = filterflags_general_tags(tags_list, **kwargs)
        >>> print(flags)
        >>> result = ut.compress(tags_list, flags)
        >>> print('result = %r' % (result,))

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_tags import *  # NOQA
        >>> import utool as ut
        >>> tags_list = [['v'], [], ['P'], ['P'], ['n', 'o'], [], ['n', 'N'], ['e', 'i', 'p', 'b', 'n'], ['n'], ['n'], ['N']]
        >>> has_all = 'n'
        >>> min_num = 1
        >>> flags = filterflags_general_tags(tags_list, has_all=has_all, min_num=min_num)
        >>> result = ut.compress(tags_list, flags)
        >>> print('result = %r' % (result,))

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_tags import *  # NOQA
        >>> import utool as ut
        >>> tags_list = [['vn'], ['vn', 'no'], ['P'], ['P'], ['n', 'o'], [], ['n', 'N'], ['e', 'i', 'p', 'b', 'n'], ['n'], ['n', 'nP'], ['NP']]
        >>> kwargs = {
        >>>     'any_endswith': 'n',
        >>>     'any_match': None,
        >>>     'any_startswith': 'n',
        >>>     'has_all': None,
        >>>     'has_any': None,
        >>>     'has_none': None,
        >>>     'max_num': 3,
        >>>     'min_num': 1,
        >>>     'none_match': ['P'],
        >>> }
        >>> flags = filterflags_general_tags(tags_list, **kwargs)
        >>> filtered = ut.compress(tags_list, flags)
        >>> result = ('result = %s' % (ut.repr2(filtered),))
        result = [['vn', 'no'], ['n', 'o'], ['n', 'N'], ['n'], ['n', 'nP']]
    """
    import numpy as np
    import utool as ut

    def _fix_tags(tags):
        if ignore_case:
            return set([]) if tags is None else {six.text_type(t.lower()) for t in tags}
        else:
            return set([]) if tags is None else {six.text_type() for t in tags}

    if logic is None:
        logic = 'and'

    logic_func = {
        'and': np.logical_and,
        'or': np.logical_or,
    }[logic]

    default_func = {
        'and': np.ones,
        'or': np.zeros,
    }[logic]

    tags_list_ = [_fix_tags(tags_) for tags_ in tags_list]
    flags = default_func(len(tags_list_), dtype=np.bool)

    if min_num is not None:
        flags_ = [len(tags_) >= min_num for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    if max_num is not None:
        flags_ = [len(tags_) <= max_num for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    if has_any is not None:
        has_any = _fix_tags(set(ut.ensure_iterable(has_any)))
        flags_ = [len(has_any.intersection(tags_)) > 0 for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    if has_none is not None:
        has_none = _fix_tags(set(ut.ensure_iterable(has_none)))
        flags_ = [len(has_none.intersection(tags_)) == 0 for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    if has_all is not None:
        has_all = _fix_tags(set(ut.ensure_iterable(has_all)))
        flags_ = [len(has_all.intersection(tags_)) == len(has_all) for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    def _test_item(tags_, fields, op, compare):
        t_flags = [any([compare(t, f) for f in fields]) for t in tags_]
        num_passed = sum(t_flags)
        flag = op(num_passed, 0)
        return flag

    def _flag_tags(tags_list, fields, op, compare):
        flags = [_test_item(tags_, fields, op, compare) for tags_ in tags_list_]
        return flags

    def _exec_filter(flags, tags_list, fields, op, compare):
        if fields is not None:
            fields = ut.ensure_iterable(fields)
            if ignore_case:
                fields = [f.lower() for f in fields]
            flags_ = _flag_tags(tags_list, fields, op, compare)
            logic_func(flags, flags_, out=flags)
        return flags

    flags = _exec_filter(
        flags, tags_list, any_startswith,
        operator.gt, six.text_type.startswith)

    flags = _exec_filter(
        flags, tags_list, in_any,
        operator.gt, operator.contains)

    flags = _exec_filter(
        flags, tags_list, any_endswith,
        operator.gt, six.text_type.endswith)

    flags = _exec_filter(
        flags, tags_list, any_match,
        operator.gt, lambda t, f: re.match(f, t))

    flags = _exec_filter(
        flags, tags_list, none_match,
        operator.eq, lambda t, f: re.match(f, t))
    return flags


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m utool.util_tags
        python -m utool.util_tags --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""

TODO: rename to something better.
this is pretty much just a mass file
editor / inspector

currently just checks invalidness
does not correct anything yet.



TODO: Fixes/Warns about the folloing problems:
    * incompatible 2/3 names
    * no future print
    * no utool inject
    * unreference tagged urls
    * ill formated docstrings
    * bad init files
    * no setup file
    * docstrings are prefixed with r
    * tests without any asserts or want lines

    * Unused functions?


rob sp DOCTEST_ENABLE ENABLE_DOCTEST
rob sp DOCTEST_DIABLE DISABLE_DOCTEST

rob gp "python .* --allexamples"

"""
from __future__ import absolute_import, division, print_function
import utool as ut  # NOQA
import os


if 'CODE_DIR' in os.environ:
    CODE_DIR = os.environ.get('CODE_DIR')

# Non local project repos
(IBEIS_REPO_URLS, IBEIS_REPO_DIRS) = ut.repo_list([
    'https://github.com/Erotemic/utool.git',
    'https://github.com/Erotemic/guitool.git',
    'https://github.com/Erotemic/plottool.git',
    'https://github.com/Erotemic/vtool.git',
    'https://github.com/bluemellophone/detecttools.git',
    'https://github.com/Erotemic/hesaff.git',
    'https://github.com/bluemellophone/pyrf.git',
    'https://github.com/Erotemic/ibeis.git',
    # 'https://github.com/aweinstock314/cyth.git',
    #'https://github.com/hjweide/pygist',
], CODE_DIR, forcessh=False)


ut.set_code_dir(CODE_DIR)
ut.set_project_repos(IBEIS_REPO_URLS, IBEIS_REPO_DIRS)


def ensure_ibeis_control_explicit_namespace(varname_list):
    # <input>
    import ibeis
    namespace = 'const'
    module = ibeis.control.IBEISControl
    fpath = module.__file__
    # </input>
    varname_list = [
        'ENCOUNTER_TABLE', 'EG_RELATION_TABLE', 'AL_RELATION_TABLE',
        'GL_RELATION_TABLE', 'CHIP_TABLE', 'FEATURE_TABLE',
        'LBLIMAGE_TABLE', 'CONTRIBUTOR_TABLE', 'LBLTYPE_TABLE',
        'METADATA_TABLE', 'VERSIONS_TABLE'
    ]
    ensure_explicit_namespace(fpath, namespace, varname_list)


def ensure_explicit_namespace(fpath, namespace, varname_list):
    import re
    import utool as ut

    text = ut.read_from(fpath)
    orig_text = text
    new_text = text

    for varname in varname_list:
        regex = ''.join((
            ut.named_field('prefix', '[^.]'),
            ut.named_field('var', ut.whole_word(varname)),
        ))
        repl = ''.join((
            ut.bref_field('prefix'),
            namespace, '.',
            ut.bref_field('var')
        ))

        new_text = re.sub(regex, repl, new_text)

    textdiff = ut.get_textdiff(orig_text, new_text)
    print(textdiff)
    if ut.user_cmdline_prompt('Does the text look good?'):
        # if diff looks good write
        ut.write_to(fpath, new_text)


def abstract_external_module_cv2():
    from os.path import join  # NOQA
    modname = 'cv2'
    repo_dirs = ut.get_project_repo_dirs()
    #exclude_dirs = [join(dpath, 'build') for dpath in repo_dirs]

    grepkw = dict(
        #exclude_dirs=exclude_dirs,
        dpath_list=repo_dirs,
        greater_exclude_dirs=ut.get_standard_exclude_dnames(),
        recursive=True,
    )
    modregex = r'\<' + modname + r'\>'
    tup = ut.grep(modregex, **grepkw)
    fpath_list, line_list, lineno_list = tup
    fpath_list = [fpath.replace('\\', '/') for fpath in fpath_list]
    print('\n'.join(fpath_list))


def filter_comented_lines():
    pass


def change_doctestcommand_to_use_dashm_flag():
    r"""
    VimRegex: # note sure how to execute replace command in vim in one lin
    %s/python\s*\([A-Za-z_]+[\\/]\S*\)\.py\(.*\)/python -m \1 \2

    """
    # http://stackoverflow.com/questions/18737863/passing-a-function-to-re-sub-in-python
    # CANNOT USE [^ ] FOR SOME GOD DAMN REASON USE /S instead
    regex_list = ['python [A-Za-z_]+[\\/]\\S* --allexamples']
    dpath_list = [
        ut.ensure_crossplat_path(ut.truepath('~/code/utool/utool')),
        ut.ensure_crossplat_path(ut.truepath('~/code/ibeis/ibeis')),
        ut.ensure_crossplat_path(ut.truepath('~/code/vtool/vtool')),
        ut.ensure_crossplat_path(ut.truepath('~/code/plottool/plottool')),
        ut.ensure_crossplat_path(ut.truepath('~/code/guitool/guitool')),
    ]
    #ut.named_field_repl(['python ', ('modrelpath',),])
    #['python ', ('modrelpath', 'utool[\\/].*'), '--allexamples'])
    res = ut.grep(regex_list, recursive=True, dpath_list=dpath_list, verbose=True)
    found_filestr_list, found_lines_list, found_lxs_list = res
    fpath = res[0][0]

    import re
    keypat_list = [
        ('prefix', 'python\\s*'),
        ('modrelpath', '[A-Za-z_]+[\\/]\\S*'),
        ('suffix', '.*'),
    ]
    namedregex = ut.named_field_regex(keypat_list)

    # Define function to pass to re.sub
    def replmodpath(matchobj):
        groupdict_ = matchobj.groupdict()
        relpath = groupdict_['modrelpath']
        prefix = groupdict_['prefix']
        suffix = groupdict_['suffix']
        modname = relpath
        modname = modname.replace('\\', '.')
        modname = modname.replace('/', '.')
        modname = modname.replace('.py', '')
        return prefix + '-m ' + modname + suffix

    for fpath in found_filestr_list:
        text = ut.read_from(fpath)
        #matchobj = re.search(namedregex, text, flags=re.MULTILINE)
        #print(text)
        #for matchobj in re.finditer(namedregex, text):
        #    print(ut.get_match_text(matchobj))
        #    print('--')
        newtext = re.sub(namedregex, replmodpath, text)
        # Perform replacement
        ut.write_to(fpath, newtext)
        #print('\n'.join(newtext.splitlines()[-10:]))


def ensure_future_compatible(mod_fpath):
    failed_future_list = []
    # Test for print function
    futureline = '^from __future__ import absolute_import, division, print_function, unicode_literals$'
    lines, lineno = ut.grepfile(mod_fpath, futureline)
    if len(lines) == 0:
        print(mod_fpath)
        failed_future_list.append(mod_fpath)
        lines, lineno = ut.grepfile(mod_fpath, futureline)


def check_six_moves_compatibility(mod_fpath):
    six_moves = ['zip', 'map', 'range', 'filter', 'cPickle', 'cStringio', 'zip_longest']
    for funcname in six_moves:
        funcname_regex = r'\b%s\b' % (funcname)
        lines, lineno = ut.grepfile(mod_fpath, funcname_regex)
        if len(lines) > 0:
            six_import = 'from six.moves import .*' + funcname
            lines_, lineno = ut.grepfile(mod_fpath, six_import)
            if len(lines_) == 0:
                print(mod_fpath + ' failed ' + funcname)
                print(lines)
                print(lines_)


def ensure_no_invalid_commands(mod_fpath):
    command_list = ['raw_input', '__builtins__', 'xrange', 'izip_longest']
    for funcname in command_list:
        funcname_regex = r'\b%s\b' % (funcname)
        lines, lineno = ut.grepfile(mod_fpath, funcname_regex)
        if len(lines) > 0:
            print(mod_fpath + ' failed ' + funcname)
            print(lines)


def ensure_utool_compatible(mod_fpath):
    ut_inject_line1 = r'print, rrr, profile ='
    ut_inject_line2 = r'\(print, rrr, profile\) ='
    ut_inject_line3 = r'inject2\(__name__,'
    ut_inject_lines = (ut_inject_line1, ut_inject_line2, ut_inject_line3)
    #ut.inject(__name'
    lines, lineno = ut.grepfile(mod_fpath, ut_inject_lines)
    if len(lines) == 0:
        print(mod_fpath + ' does not have utool')


#def ensure_compatible_modfpath_list(mod_fpath_list):


if __name__ == '__main__':
    package_dir = ut.truepath('~/code/ibeis/ibeis')
    if 'module_fpath_list' not in vars():
        module_fpath_list = ut.glob_python_modules(package_dir)

    for mod_fpath in module_fpath_list:
        #ensure_compatible_modfpath(mod_fpath)
        #check_six_moves_compatibility(mod_fpath)
        #ensure_utool_compatible(mod_fpath)
        ensure_no_invalid_commands(mod_fpath)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import inspect
from os.path import split, splitext, dirname, basename


def get_stack_frame(N=0, strict=True):
    frame_level0 = inspect.currentframe()
    frame_cur = frame_level0
    for _ix in range(N + 1):
        frame_next = frame_cur.f_back
        if frame_next is None:
            if strict:
                raise AssertionError('Frame level %r is root' % _ix)
            else:
                break
        frame_cur = frame_next
    return frame_cur


def get_caller_lineno(N=0, strict=True):
    parent_frame = get_stack_frame(N=N + 1, strict=strict)
    lineno =  parent_frame.f_lineno
    return lineno


def get_caller_name(N=0, strict=True):
    """ Standalone version of get_caller_name """
    if isinstance(N, (list, tuple)):
        name_list = []
        for N_ in N:
            try:
                name_list.append(get_caller_name(N_))
            except AssertionError:
                name_list.append('X')
        return '[' + ']['.join(name_list) + ']'
    # <get_parent_frame>
    parent_frame = get_stack_frame(N=N + 2, strict=strict)
    # </get_parent_frame>
    caller_name = parent_frame.f_code.co_name
    if caller_name == '<module>':
        co_filename = parent_frame.f_code.co_filename
        caller_name = splitext(split(co_filename)[1])[0]
    if caller_name == '__init__':
        co_filename = parent_frame.f_code.co_filename
        caller_name = basename(dirname(co_filename)) + '.' + caller_name
    return caller_name
